#!/usr/bin/env python
#
# notebook.py - The NotebookAction class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`NotebookAction` class, an action which
starts an IPython kernel and a Jupyter notebook server, allowing the user
to interact with FSLeyes via a jupyter notebook.
"""


import               os
import os.path    as op
import subprocess as sp
import               sys
import               time
import               atexit
import               shutil
import               logging
import               binascii
import               tempfile
import               warnings
import               threading
import               contextlib
import               webbrowser

import               wx
import jinja2     as j2

import fsleyes_widgets.utils.progress as progress
import fsleyes_widgets.utils.status   as status
import fsl.utils.settings             as settings
import fsl.utils.tempdir              as tempdir
import fsl.utils.idle                 as idle

import                                   fsleyes
import fsleyes.main                   as fsleyes_main
import fsleyes.strings                as strings
import fsleyes.actions.screenshot     as screenshot

import fsleyes.actions.base           as base
import fsleyes.actions.runscript      as runscript

try:
    import                            zmq

    import zmq.eventloop.zmqstream as zmqstream

    import ipykernel.ipkernel      as ipkernel
    import ipykernel.iostream      as iostream
    import ipykernel.zmqshell      as zmqshell
    import ipykernel.heartbeat     as heartbeat

    import notebook.notebookapp    as notebookapp

    import IPython.display         as display

    import jupyter_client          as jc
    import jupyter_client.session  as jcsession

    ENABLED = True

except ImportError:

    class mock(object):
        pass

    # so the sub-class defs in
    # this module do not error
    zmqshell                     = mock()
    zmqshell.ZMQInteractiveShell = mock
    ipkernel                     = mock()
    ipkernel.IPythonKernel       = mock

    ENABLED = False


log = logging.getLogger(__name__)


class NotebookAction(base.Action):
    """The ``NotebookAction`` is an :class:`.Action` which (if necessary)
    starts an embedded IPython kernel and a jupyter notebook server, and
    opens the server home page in a web browser allowing the user to interact
    with FSLeyes via notebooks.
    """


    def __init__(self, overlayList, displayCtx, frame):
        """Create a ``NotebookAction``.

        :arg overlayList: The :class:`.OverlayList`.
        :arg displayCtx:  The master :class:`.DisplayContext`.
        :arg frame:       The :class:`.FSLeyesFrame`.
        """
        base.Action.__init__(
            self, overlayList, displayCtx, self.__openNotebooks)

        # permanently disable if any
        # dependencies are not present
        self.enabled  = ENABLED
        self.__frame  = frame
        self.__kernel = None
        self.__server = None


    def __openNotebooks(self, nbfile=None):
        """Called when this ``NotebookAction`` is invoked. Starts the
        server and kernel if necessary.

        If the server/kernel have not yet been started and ``nbfile`` is
        provided, the server will be started with ``nbfile`` opened.
        """

        # If the kernel and server are both
        # ok, open the notebook homepage
        if self.__kernel is not None and \
           self.__server is not None and \
           self.__kernel.is_alive()  and \
           self.__server.is_alive():
            webbrowser.open(self.__server.url)
            return

        # have the kernel or server threads crashed?
        if self.__kernel is not None and not self.__kernel.is_alive():
            self.__kernel = None
        if self.__server is not None and not self.__server.is_alive():
            self.__server = None

        # show a progress dialog if we need
        # to initialise the kernel or server
        if self.__kernel is None or self.__server is None:
            title   = strings.titles[self, 'init']
            progdlg = progress.Bounce(title)
        else:
            progdlg = None

        try:
            # start the kernel/server, and show
            # an error if something goes wrong
            errt = strings.titles[  self, 'init.error']
            errm = strings.messages[self, 'init.error']
            with status.reportIfError(errt, errm):
                if self.__kernel is None:
                    self.__kernel = self.__startKernel(progdlg)
                if self.__server is None:
                    self.__server = self.__startServer(progdlg, nbfile)

        finally:
            if progdlg is not None:
                progdlg.Destroy()
                progdlg = None


    def __bounce(self, secs, progdlg):
        """Used by :meth:`__startKernel` and :meth:`__startServer`. Blocks for
        ``secs``, bouncing the :class:`.Progress` dialog ten times per second,
        and yielding to the ``wx`` main loop.
        """
        for i in range(int(secs * 10)):
            progdlg.DoBounce()
            wx.GetApp().Yield()
            time.sleep(0.1)


    def __startKernel(self, progdlg):
        """Attempts to create and start a :class:`BackgroundIPythonKernel`.

        :returns: the kernel if it was started.

        :raises: A :exc:`RuntimeError` if the kernel did not start.
        """
        progdlg.UpdateMessage(strings.messages[self, 'init.kernel'])
        kernel = BackgroundIPythonKernel(
            self.overlayList,
            self.displayCtx,
            self.__frame)
        kernel.start()
        self.__bounce(2, progdlg)

        if not kernel.is_alive():
            raise RuntimeError('Could not start IPython kernel: '
                               '{}'.format(kernel.error))

        return kernel


    def __startServer(self, progdlg, nbfile=None):
        """Attempts to create and start a :class:`NotebookServer`.

        :returns: the server if it was started.

        :raises: A :exc:`RuntimeError` if the serer did not start.
        """

        progdlg.UpdateMessage(strings.messages[self, 'init.server'])
        server = NotebookServer(self.__kernel.connfile, nbfile)
        server.start()

        elapsed = 0

        while elapsed < 5 and not server.ready:
            self.__bounce(0.5, progdlg)
            elapsed += 0.5

        if elapsed >= 5 or not server.is_alive():
            raise RuntimeError('Could not start notebook server: '
                               '{}'.format(server.stderr))

        return server


class BackgroundIPythonKernel:
    """The BackgroundIPythonKernel creates an IPython jupyter kernel and makes
    it accessible over tcp (on the local machine only).

    Before FSLeyes 0.28.0, this class derived from ``threading.Thread``, and
    ran its own ``IOLoop`` which was used to run the IPython kernel.  But now
    the IPython kernel is run on the main thread, by repeatedly scheduling
    calls to ``IPythonKernel.do_one_iteration`` on :func:`.idle.idle`.

    This update was necessary due to changes in ``ipykernel`` from version 4
    to version 5, namely that ipykernel 5 uses ``asyncio`` co-routines, making
    it difficult to transfer execution between threads.

    Because the ``BackgroundIPythonKernel`` used to be a thread, it is still
    necessary to call its :meth:`start` method to start the kernel loop.

    The Jupyter/IPython documentation is quite fragmented at this point in
    time; `this github issue <https://github.com/ipython/ipython/issues/8097>`_
    was useful in figuring out the implementation details.
    """


    def __init__(self, overlayList, displayCtx, frame):
        """Set up the kernel and ``zmq`` ports. A jupyter connection file
        containing the information needed to connect to the kernel is saved
        to a temporary file - its path is accessed as an attribute
        called :meth:`connfile`.

        :arg overlayList: The :class:`.OverlayList`.
        :arg displayCtx:  The master :class:`.DisplayContext`.
        :arg frame:       The :class:`.FSLeyesFrame`.
        """

        ip                 = '127.0.0.1'
        transport          = 'tcp'
        addr               = '{}://{}'.format(transport, ip)
        self.__connfile    = None
        self.__kernel      = None
        self.__error       = None
        self.__lastIter    = 0
        self.__overlayList = overlayList
        self.__displayCtx  = displayCtx
        self.__frame       = frame
        self.__env         = runscript.fsleyesScriptEnvironment(
            frame,
            overlayList,
            displayCtx)[1]

        self.__env['screenshot'] = self.__screenshot

        with warnings.catch_warnings():
            warnings.simplefilter('ignore', category=DeprecationWarning)

            # Use an empty key to disable message signing
            session = jcsession.Session(key=b'')
            context = zmq.Context.instance()

            # create sockets for kernel communication
            shellsock   = context.socket(zmq.ROUTER)
            stdinsock   = context.socket(zmq.ROUTER)
            controlsock = context.socket(zmq.ROUTER)
            iopubsock   = context.socket(zmq.PUB)

            shellstrm   = zmqstream.ZMQStream(shellsock)
            controlstrm = zmqstream.ZMQStream(controlsock)

            # I/O and heartbeat communication
            # are managed by separate threads.
            self.__iopub     = iostream.IOPubThread(iopubsock)
            self.__heartbeat = heartbeat.Heartbeat(zmq.Context(),
                                                   (transport, ip, 0))
            iopubsock = self.__iopub.background_socket

            self.__heartbeat.start()
            self.__iopub.start()

            # Streams which redirect stdout/
            # stderr to the iopub socket
            stdout = iostream.OutStream(session, self.__iopub, u'stdout')
            stderr = iostream.OutStream(session, self.__iopub, u'stderr')

            # TCP ports for all sockets
            shellport   = shellsock  .bind_to_random_port(addr)
            stdinport   = stdinsock  .bind_to_random_port(addr)
            controlport = controlsock.bind_to_random_port(addr)
            iopubport   = iopubsock  .bind_to_random_port(addr)
            hbport      = self.__heartbeat.port

            # Create the kernel
            self.__kernel = FSLeyesIPythonKernel.instance(
                stdout,
                stderr,
                shell_class=FSLeyesIPythonShell,
                session=session,
                shell_streams=[shellstrm, controlstrm],
                iopub_socket=iopubsock,
                stdin_socket=stdinsock,
                user_ns=self.__env,
                log=logging.getLogger('ipykernel.kernelbase'))

        # write connection file to a temp dir
        hd, fname = tempfile.mkstemp(
            prefix='fsleyes-kernel-{}.json'.format(os.getpid()),
            suffix='.json')
        os.close(hd)

        self.__connfile = fname

        log.debug('IPython kernel connection file: %s', fname)

        jc.write_connection_file(
            fname,
            shell_port=shellport,
            stdin_port=stdinport,
            iopub_port=iopubport,
            control_port=controlport,
            hb_port=hbport,
            ip=ip)

        atexit.register(os.remove, self.__connfile)


    def is_alive(self):
        """Returns ``True`` if the kernel loop appears to be running, ``False``
        otherwise
        """
        # We check that the last kernel
        # iteration was not too long ago,
        # that an error hasn't occurred,
        # and that the worker threads are
        # alive.
        now   = time.time()
        delta = 5.0

        return (now - self.__lastIter) < delta and \
               self.__error is None            and \
               self.__iopub.thread.is_alive()  and \
               self.__heartbeat.is_alive()


    @property
    def kernel(self):
        """The ``IPythonKernel`` object. """
        return self.__kernel


    @property
    def env(self):
        """The namespace passed to the kernel."""
        return self.__env


    @property
    def connfile(self):
        """The jupyter connection file containing information to connect
        to the IPython kernel.
        """
        return self.__connfile


    @property
    def error(self):
        """If an error occurs on the kernel loop causing it to crash, a
        reference to the ``Exception`` is stored here.
        """
        return self.__error


    def __screenshot(self, view=None):
        """Insert a screenshot of the given view panel into the notebook.
        If ``view`` is not specified, the first view is assumed.
        """
        if view is None:
            view = self.__frame.viewPanels[0]
        with tempdir.tempdir():
            screenshot.screenshot(view, 'screenshot.png')
            return display.Image('screenshot.png')


    def start(self):
        """Start the IPython kernel loop. This method returns immediately - the
        kernel loop is driven on :func:`.idle.idle`.
        """
        self.__kernel.start()
        idle.idle(self.__eventloop, after=self.__kernel._poll_interval)


    def __eventloop(self):
        """Event loop used for the IPython kernel. Calls :meth:`__kernelDispatch`,
        then schedules a future call to ``__eventloop`` via :func:`.idle.idle`
        loop.
        """
        self.__kernelDispatch()
        idle.idle(self.__eventloop, after=self.__kernel._poll_interval)


    def __kernelDispatch(self):
        """Execute one kernel iteration, by scheduling a call to
        ``IPythonKernel.do_one_iteration`` on the kernel's io loop.
        """
        try:
            loop = self.__kernel.io_loop
            loop.run_sync(self.__kernel.do_one_iteration)

            # save the time on each iteration,
            # so the is_alive method has an
            # idea of whether the kernel is
            # still running
            self.__lastIter = time.time()

        except Exception as e:
            self.__error = e
            raise


class FSLeyesIPythonKernel(ipkernel.IPythonKernel):
    """Custom IPython kernel used by FSLeyes. All this class does is ensure
    that the ``sys.stdout`` and ``sys.stderr`` streams are set appropriately
    when the kernel is executing code.
    """

    def __init__(self, stdout, stderr, *args, **kwargs):
        self.__stdout = stdout
        self.__stderr = stderr
        super(FSLeyesIPythonKernel, self).__init__(*args, **kwargs)

    @contextlib.contextmanager
    def __patch_streams(self):

        stdout     = sys.stdout
        stderr     = sys.stderr
        sys.stdout = self.__stdout
        sys.stderr = self.__stderr

        try:
            yield
        finally:

            sys.stdout = stdout
            sys.stderr = stderr
            self.__stdout.flush()
            self.__stderr.flush()

    def execute_request(self, *args, **kwargs):
        with self.__patch_streams():
            return super(FSLeyesIPythonKernel, self).execute_request(
                *args, **kwargs)

    def dispatch_control(self, *args, **kwargs):
        with self.__patch_streams():
            return super(FSLeyesIPythonKernel, self).dispatch_control(
                *args, **kwargs)

    def dispatch_shell(self, *args, **kwargs):
        with self.__patch_streams():
            return super(FSLeyesIPythonKernel, self).dispatch_shell(
                *args, **kwargs)


class FSLeyesIPythonShell(zmqshell.ZMQInteractiveShell):
    """Custom IPython shell class used by FSLeyes. """

    def enable_gui(self, gui):
        """Overrides  ``ipykernel.zmqshell.ZMQInteractiveShell.enable_gui``.

        The default implementation will attempt to change the IPython GUI
        integration event loop, which will conflict with our own event loop
        in the :class:`BackgroundIPythonKernel`. So this implementation
        does nothing.
        """
        pass


class NotebookServer(threading.Thread):
    """Thread which starts a jupyter notebook server, and waits until
    it dies or is killed.

    The server is configured such that all notebooks will connect to
    the same kernel, specified by the ``kernelFile`` parameter to
    :meth:`__init__`.
    """


    def __init__(self, connfile, nbfile=None):
        """Create a ``NotebookServer`` thread.

        :arg connfile: Connection file of the IPython kernel to connect to.
        :arg nbfile:   Path to a notebook file which should be opened on
                       startup.
        """

        threading.Thread.__init__(self)
        self.daemon        = True
        self.__connfile    = connfile
        self.__nbfile      = nbfile
        self.__stdout      = None
        self.__stderr      = None
        self.__port        = None
        self.__token       = binascii.hexlify(os.urandom(24)).decode('ascii')


    @property
    def port(self):
        """Returns the TCP port that the notebook server is listening on.
        Will return ``None`` before the server has started.
        """
        if self.__port is not None:
            return self.__port

        self.__port = self.__readPort()
        return self.__port


    def __readPort(self):
        for server in notebookapp.list_running_servers():
            if server['token'] == self.__token:
                return server['port']
        return None


    @property
    def ready(self):
        """Returns ``True`` if the server is running and ready, ``False``
        otherwise.
        """
        return self.__readPort() is not None


    @property
    def token(self):
        """Returns an authentication token to use for connectng to the
        notebook server.
        """
        return self.__token


    @property
    def url(self):
        """Returns the URL to use to connect to this server. """
        return 'http://localhost:{}?token={}'.format(self.port, self.token)


    @property
    def stdout(self):
        """After the server has died, returns its standard output. While the
        server is still running, returns ``None``.
        """
        return self.__stdout


    @property
    def stderr(self):
        """After the server has died, returns its standard error. While the
        server is still running, returns ``None``.
        """
        return self.__stderr


    def run(self):
        """Sets up a server configuration file, and then calls
        ``jupyer-notebook`` via ``subprocess.Popen``. Waits until
        the notebook process dies.
        """

        # copy our custom jupyter config template
        # directory to a temporary location that
        # will be deleted on exit.
        cfgdir = op.join(tempfile.mkdtemp(prefix='fsleyes-jupyter'), 'config')
        shutil.copytree(op.join(fsleyes.assetDir, 'assets', 'jupyter'), cfgdir)

        log.debug('Copied notebook configuration to %s', cfgdir)

        self.__initConfigDir(cfgdir)

        # Set up the environment in which the
        # server will run - make sure FSLeyes
        # is on the PYTHONPATH, and
        # JUPYTER_CONFIG_DIR is set, so our
        # custom bits and pieces will be found.
        env        = dict(os.environ)
        pythonpath = os.pathsep.join((cfgdir, env.get('PYTHONPATH', '')))

        env['JUPYTER_CONFIG_DIR'] = cfgdir
        env['PYTHONPATH']         = pythonpath

        # command to start the notebook server
        # in a sub-process - we run the server
        # via a wrapper function called nbmain,
        # defined below.
        cmd = [sys.executable, fsleyes_main.__file__]

        # py2app manipulates the PYTHONPATH,
        # so we pass the cfgdir through as a
        # command-line argument - it is picked
        # up again by the nbmain function.
        cmd.extend(('notebook', 'server', cfgdir))

        if self.__nbfile is not None:
            cmd.append(self.__nbfile)

        log.debug('Running notebook server via %s notebook', cmd[0])
        self.__nbproc = sp.Popen(cmd,
                                 stdout=sp.PIPE,
                                 stderr=sp.PIPE,
                                 env=env)

        def killServer():
            if self.__nbproc is not None:
                # We need two CTRL+Cs to kill
                # the notebook server
                self.__nbproc.terminate()
                self.__nbproc.terminate()

                # Give it a little time, then
                # force kill if needed
                try:
                    self.__nbproc.wait(0.25)
                except sp.TimeoutExpired:
                    self.__nbproc.kill()
                self.__nbproc = None

            shutil.rmtree(op.abspath(op.join(cfgdir, '..')))

        # kill the server when we get killed
        atexit.register(killServer)

        # wait forever
        o, e          = self.__nbproc.communicate()
        self.__stdout = o.decode()
        self.__stderr = e.decode()
        self.__nbproc = None

        # if we've gotten this far,
        # call our atexit handler
        # directly, and unregister
        # it
        killServer()
        atexit.unregister(killServer)


    def __initConfigDir(self, cfgdir):
        """Creates a copy of the FSLeyes ``/assets/jupyter/`` configuration
        directory in ``$TMPDIR``, and customises its settings accordingly.
        """

        nbextdir    = op.join(cfgdir, 'nbextensions')
        defaultPort = settings.read('fsleyes.notebook.port', 8888)

        # Environment for generating a jupyter
        # notebook server configuration file
        cfgenv = {
            'fsleyes_nbserver_port'       : defaultPort,
            'fsleyes_nbserver_token'      : self.__token,
            'fsleyes_nbserver_dir'        : os.getcwd(),
            'fsleyes_nbserver_static_dir' : cfgdir,
            'fsleyes_nbextension_dir'     : nbextdir,
            'fsleyes_kernel_connfile'     : self.__connfile,
        }

        with open(op.join(nbextdir, 'fsleyes_notebook_intro.md'), 'rt') as f:
            intro = f.read()

        # Environment for generating the
        # notebook template extension
        extenv = {
            'intro' : intro.replace('\n', '\\n'),
        }

        cfgfile = op.join(cfgdir, 'jupyter_notebook_config.py')
        extfile = op.join(cfgdir, 'nbextensions', 'notebook_template.js')

        files = [cfgfile, extfile]
        envs  = [cfgenv,  extenv]

        for fn, e in zip(files, envs):
            with open(fn, 'rt') as f: template = j2.Template(f.read())
            with open(fn, 'wt') as f: f.write(template.render(**e))


def nbmain(argv):
    """Wrapper around a Jupyter Notebook server entry point. Invoked by the
    :class:`NotebookServer`, via a hook in :func:`fsleyes.main.main`.
    """

    if argv[:2] != ['notebook', 'server']:
        raise RuntimeError('argv does not look like notebook main arguments '
                           '(first args are not \'notebook server\')')

    argv = argv[2:]

    # run the notebook server
    from notebook.notebookapp import main as nbmain

    # first argument is a path
    # to add to the PYTHONPATH.
    # See NotebookServer.run.
    sys.path.insert(0, argv[0])

    # remaining arguments are passed
    # through to notebookapp.main
    return nbmain(argv=argv[1:])



#!/usr/bin/env python
#
# scene3dtoolbar.py - The Scene3DToolBar class
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`Scene3DToolBar` class, a FSLeyes control
which implements a toolbar for use with the :class:`.Scene3DPanel`.
"""


import wx


import fsleyes_props                 as props
import fsleyes.controls.controlpanel as ctrlpanel
import fsleyes.views.scene3dpanel    as scene3dpanel
import fsleyes.toolbar               as fsltoolbar
import fsleyes.strings               as strings
import fsleyes.actions               as actions
import fsleyes.icons                 as fslicons
import fsleyes.tooltips              as fsltooltips



class Scene3DToolBar(ctrlpanel.ControlToolBar):
    """The ``Scene3DToolBar`` is a FSLeyes control which implements a toolbar for
    use with the :class:`.Scene3DPanel`.
    """


    showCursorAndLegend = props.Boolean(default=True)
    """This property is linked to a button on the toolbar which allows the
    user to simultaneously toggle the :attr:`.SceneOpts.showCursor` and
    :attr:`.Scene3DOpts.showLegend` properties.
    """


    @staticmethod
    def supportedViews():
        """The ``Scene3DToolBar`` is restricted for use with the
        :class:`.Scene3DPanel`.
        """
        return [scene3dpanel.Scene3DPanel]


    def __init__(self, parent, overlayList, displayCtx, panel):
        """Create a ``Scene3DToolBar``.

        :arg parent:      The :mod:`wx` parent object.
        :arg overlayList: The :class:`.OverlayList` instance.
        :arg displayCtx:  The :class:`.DisplayContext` instance.
        :arg panel:       The :class:`.Scene3DPanel` instance.
        """

        ctrlpanel.ControlToolBar.__init__(self,
                                          parent,
                                          overlayList,
                                          displayCtx,
                                          panel,
                                          height=24,
                                          kbFocus=True)

        self.panel = panel

        self.addListener('showCursorAndLegend',
                         self.name,
                         self.__showCursorAndLegendChanged)

        self.__makeTools()


    def destroy(self):
        """Must be called when this ``Scene3DToolBar`` is no longer in use.
        Removes some property listeners, and calls the base class
        implementation.
        """
        self.removeListener('showCursorAndLegend', self.name)

        ctrlpanel.ControlToolBar.destroy(self)


    def __makeTools(self):
        """Called by :meth:`__init__`. Creates the toolbar widgets. """

        panel   = self.panel
        opts    = panel.sceneOpts
        profile = panel.currentProfile

        icons = {
            'screenshot'          : fslicons.findImageFile('camera24'),
            'resetDisplay'        : fslicons.findImageFile('resetZoom24'),
            'showCursorAndLegend' : [
                fslicons.findImageFile('addHighlight24'),
                fslicons.findImageFile('add24')],
            'movieMode'        : [
                fslicons.findImageFile('movieHighlight24'),
                fslicons.findImageFile('movie24')],
            'CanvasSettingsPanel' : [
                fslicons.findImageFile('spannerHighlight24'),
                fslicons.findImageFile('spanner24')],
        }

        tooltips = {
            'screenshot'   : fsltooltips.actions[   panel,     'screenshot'],
            'resetDisplay' : fsltooltips.actions[   profile,   'resetDisplay'],
            'movieMode'    : fsltooltips.properties[panel,     'movieMode'],
            'showCursorAndLegend' : fsltooltips.properties[
                self, 'showCursorAndLegend'],
            'zoom'         : fsltooltips.properties[opts, 'zoom'],
            'CanvasSettingsPanel' : fsltooltips.actions[
                panel, 'CanvasSettingsPanel'],
        }

        targets = {
            'screenshot'          : panel,
            'resetDisplay'        : profile,
            'movieMode'           : panel,
            'showCursorAndLegend' : self,
            'zoom'                : opts,
            'CanvasSettingsPanel' : panel,
        }

        toolSpecs = [
            actions.ToggleActionButton(
                'CanvasSettingsPanel',
                actionKwargs={'floatPane' : True},
                icon=icons['CanvasSettingsPanel'],
                tooltip=tooltips['CanvasSettingsPanel']),
            actions.ActionButton('screenshot',
                                 icon=icons['screenshot'],
                                 tooltip=tooltips['screenshot']),
            'div',
            props  .Widget(      'movieMode',
                                 icon=icons['movieMode'],
                                 tooltip=tooltips['movieMode']),
            props  .Widget(      'showCursorAndLegend',
                                 icon=icons['showCursorAndLegend'],
                                 tooltip=tooltips['showCursorAndLegend']),
            actions.ActionButton('resetDisplay',
                                 icon=icons['resetDisplay'],
                                 tooltip=tooltips['resetDisplay']),
            props.Widget(        'zoom',
                                 spin=True,
                                 slider=True,
                                 showLimits=False,
                                 spinWidth=5,
                                 tooltip=tooltips['zoom']),
        ]

        tools = []
        nav   = []

        for spec in toolSpecs:

            if spec == 'div':
                tools.append(fsltoolbar.ToolBarDivider(self,
                                                       height=24,
                                                       orient=wx.VERTICAL))
                continue

            widget    = props.buildGUI(self, targets[spec.key], spec)
            navWidget = widget

            if spec.key in ('zoom', ):
                widget = self.MakeLabelledTool(
                    widget,
                    strings.properties[targets[spec.key], spec.key])

            tools.append(widget)
            nav  .append(navWidget)

        self.SetTools(tools, destroy=True)
        self.setNavOrder(nav)


    def __showCursorAndLegendChanged(self, *a):
        """Called when the :attr:`showCursorAndLegend` property is changed.
        Propagates the change on to the :attr:`.Scene3DOpts.showCursor` and
        :attr:`.Scene3DOpts.showLegend` properties.
        """

        opts            = self.panel.sceneOpts
        opts.showCursor = self.showCursorAndLegend
        opts.showLegend = self.showCursorAndLegend



#!/usr/bin/env python
#
# colourmaptexture.py - The ColourMapTexture class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`ColourMapTexture` class, a 1D
:class:`.Texture` which can be used to store a RGBA colour map.
"""


import logging
import collections.abc as abc

import numpy     as np
import OpenGL.GL as gl

from . import texture


log = logging.getLogger(__name__)


class ColourMapTexture(texture.Texture):
    """The ``ColourMapTexture`` class is a :class:`.Texture` which stores
    a RGB or RGBA colour map.

    A ``ColourMapTexture`` maps a data range to to a colour map. The data
    range may be specified by the :meth:`setDisplayRange` method, and the
    colour map by the :meth:`setColourMap` method. Alternately, both can
    be specified with the :meth:`set` method.


    In OpenGL, textures are indexed with a number between 0.0 and 1.0. So
    in order to map the data range to texture coordinates, an offset/scale
    transformation must be applied to data values. The ``ColourMapTexture``
    calculates this transformation, and makes it available via the
    :meth:`getCoordinateTransform` method.


    The colour map itself can be specified in a number of ways:

      - A ``numpy`` array of size :math:`N\\times 3` or :math:`N\\times 4`,
        containing RGB or RGBA colour values, with colour values in the range
        ``[0, 1]``.

      - A function which accepts an array of values in the range ``[0, 1]``,
        and returns an array of size :math:`N\\times 3` or :math:`N\\times 4`,
        specifying the RGB/RGBA colours that correspond to the input values.


    Some other methods are provided, for configuring the colour map:

    .. autosummary::
       :nosignatures:

       setAlpha
       setInvert
       setResolution
       setGamma
       setInterp
       setBorder
    """


    def __init__(self, name):
        """Create a ``ColourMapTexture``.

        :arg name: A unique name for this ``ColourMapTexture``.
        """

        self.__resolution   = None
        self.__cmap         = None
        self.__invert       = False
        self.__interp       = None
        self.__gamma        = None
        self.__alpha        = None
        self.__displayRange = None
        self.__border       = None
        self.__coordXform   = None

        texture.Texture.__init__(self, name, 1, 4)


    def setColourMap(self, cmap):
        """Set the colour map stored by the ``ColourMapTexture``.

        :arg cmap: The colour map, either a ``numpy`` array of size
                   :math:`N\\times 3` or :math:`N\\times 4`, specifying
                   RGB/RGBA colours, or a function which accepts values
                   in the range ``[0, 1]``, and generates corresponding
                   RGB/RGBA colours.
        """
        self.set(cmap=cmap)


    def setResolution(self, res):
        """Set the resolution (number of colours) of this ``ColourMapTexture``.
        This setting is only applicable when the colour map is specified as a
        function (see :meth:`setColourMap`).
        """
        self.set(resolution=res)


    def setAlpha(self, alpha):
        """Set the transparency of all colours in the colour map. This setting
        is only applicable when the colour map is specified as RGB values.
        """
        self.set(alpha=alpha)


    def setInvert(self, invert):
        """Invert the values in the colour map. """
        self.set(invert=invert)


    def setGamma(self, gamma):
        """Gamma correction - uses ``gamma`` as an exponent to weight the
        colour map towards the low or high end. Only applied if the
        colour map (see :meth:`setColourMap`) is specified as a function.
        """
        self.set(gamma=gamma)


    def setInterp(self, interp):
        """Set the interpolation used by this ``ColourMapTexture`` - either
        ``GL_NEAREST`` or ``GL_LINEAR``.
        """
        self.set(interp=interp)


    def setDisplayRange(self, drange):
        """Set the data range which corresponds to the colours stored in this
        ``ColourMapTexture``. A matrix which transforms values from from this
        data range into texture coordinates is available via the
        :meth:`getCoordinateTransform` method.
        """
        self.set(displayRange=drange)


    def setBorder(self, border):
        """Set the texture border colour. If ``None``, the edge colours of the
        colour map are used as the border.
        """
        self.set(border=border)


    def getCoordinateTransform(self):
        """Returns a matrix which transforms values from from the colour map
        data range (see :meth:`setDisplayRange`) into texture coordinates.
        """
        return self.__coordXform


    def set(self, **kwargs):
        """Set any parameters on this ``ColourMapTexture``. Valid keyword
        arguments are:

        ================ ============================
        ``cmap``         See :meth:`setColourMap`.
        ``invert``       See :meth:`setInvert`.
        ``interp``       See :meth:`setInterp`.
        ``alpha``        See :meth:`setAlpha`.
        ``resolution``   See :meth:`setResolution`.
        ``gamma``        See :meth:`setGamma`.
        ``displayRange`` See :meth:`setDisplayRange`.
        ``border``       See :meth:`setBorder`.
        ================ ============================
        """

        # None is a valid value for any attributes,
        # so we are using 'self' to test whether
        # or not an attribute value was passed in
        cmap         = kwargs.get('cmap',         self)
        invert       = kwargs.get('invert',       self)
        interp       = kwargs.get('interp',       self)
        alpha        = kwargs.get('alpha',        self)
        resolution   = kwargs.get('resolution',   self)
        gamma        = kwargs.get('gamma',        self)
        displayRange = kwargs.get('displayRange', self)
        border       = kwargs.get('border',       self)

        if cmap         is not self: self.__cmap         = cmap
        if invert       is not self: self.__invert       = invert
        if interp       is not self: self.__interp       = interp
        if alpha        is not self: self.__alpha        = alpha
        if displayRange is not self: self.__displayRange = displayRange
        if border       is not self: self.__border       = border
        if resolution   is not self: self.__resolution   = resolution
        if gamma        is not self: self.__gamma        = gamma

        self.__refresh()


    def __prepareTextureSettings(self):
        """Called by :meth:`__refresh`. Prepares all of the texture settings,
        and returns a tuple containing:

          - An array containing the colour map data
          - The display range
          - The interpolation setting
          - The border colour
        """

        import matplotlib.colors as colors

        alpha  = self.__alpha
        cmap   = self.__cmap
        drange = self.__displayRange
        invert = self.__invert
        interp = self.__interp
        res    = self.__resolution
        gamma  = self.__gamma
        border = self.__border

        if drange is None: drange = [0.0, 1.0]
        if invert is None: invert = False
        if interp is None: interp = gl.GL_NEAREST
        if cmap   is None: cmap   = np.zeros((4, 4), dtype=np.float32)
        if res    is None: res    = 256
        if gamma  is None: gamma  = 1

        # The fsleyes.colourmaps module creates
        # ListedColormap instances. If the given
        # cmap is one of these, there's no point
        # in using a resolution greater than the
        # number of colours in the cmap.
        if isinstance(cmap, colors.ListedColormap):
            res = min(res, cmap.colors.shape[0])

        # If cmap is a function, assume that it accepts
        # one or more scalar values between 0 and 1,
        # and converts said values into a numpy array
        # containing RGB/RGBA colours.  Crop the RGB
        # colours, as global alpha takes precedence
        if isinstance(cmap, abc.Callable):

            # Apply gamma scaling to weight
            # towards one end of the colour map
            idxs = np.linspace(0.0, 1.0, res) ** gamma
            cmap = cmap(idxs)
            cmap = cmap[:, :3]

        # If RGB, turn into RGBA. If an RGBA cmap
        # has been provided, their alpha values take
        # precedence over the global alpha setting
        if cmap.shape[1] == 3:
            newCmap        = np.ones((cmap.shape[0], 4), dtype=np.float32)
            newCmap[:, :3] = cmap
            cmap           = newCmap

            # Apply the global alpha if provided
            if alpha is not None:
                cmap[:, 3] = alpha

        # Reverse the colours if necessray
        if invert:
            cmap = cmap[::-1, :]

        # If border is provided and is
        # RGB, convert it to RGBA
        if border is not None and len(border) == 3:
            newBorder     = np.ones(4, dtype=np.float32)
            newBorder[:3] = border
            if alpha is not None:
                newBorder[3] = alpha

            border = newBorder

        return cmap, drange, interp, border


    def __refresh(self):
        """Called when any settings of this ``ColourMapTexture`` are changed.
        Re-configures the texture.
        """

        cmap, drange, interp, border = self.__prepareTextureSettings()

        imin, imax = drange

        # This transformation is used to transform input values
        # from their native range to the range [0.0, 1.0], which
        # is required for texture colour lookup. Values below
        # or above the current display range will be mapped
        # to texture coordinate values less than 0.0 or greater
        # than 1.0 respectively.
        if imax == imin: scale = 0.000000000001
        else:            scale = imax - imin

        coordXform = np.identity(4, dtype=np.float64)
        coordXform[0, 0] = 1.0 / scale
        coordXform[0, 3] = -imin * coordXform[0, 0]

        self.__coordXform = coordXform

        # The colour data is stored on
        # the GPU as 8 bit rgba tuples
        cmap = np.floor(cmap * 255)
        cmap = np.array(cmap, dtype=np.uint8)
        cmap = cmap.ravel(order='C')

        # GL texture creation stuff
        self.bindTexture()

        if border is not None:

            gl.glTexParameterfv(gl.GL_TEXTURE_1D,
                                gl.GL_TEXTURE_BORDER_COLOR,
                                border)
            gl.glTexParameteri( gl.GL_TEXTURE_1D,
                                gl.GL_TEXTURE_WRAP_S,
                                gl.GL_CLAMP_TO_BORDER)
        else:
            gl.glTexParameteri(gl.GL_TEXTURE_1D,
                               gl.GL_TEXTURE_WRAP_S,
                               gl.GL_CLAMP_TO_EDGE)

        gl.glTexParameteri(gl.GL_TEXTURE_1D,
                           gl.GL_TEXTURE_MAG_FILTER,
                           interp)
        gl.glTexParameteri(gl.GL_TEXTURE_1D,
                           gl.GL_TEXTURE_MIN_FILTER,
                           interp)

        gl.glTexImage1D(gl.GL_TEXTURE_1D,
                        0,
                        gl.GL_RGBA8,
                        len(cmap) / 4,
                        0,
                        gl.GL_RGBA,
                        gl.GL_UNSIGNED_BYTE,
                        cmap)
        self.unbindTexture()



#!/usr/bin/env python
#
# edittransform.py - The EditTransformPanel class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`EditTransformPanel` (a.k.a. "Nudge")
class, a FSLeyes control panel which allows the user to adjust the
``voxToWorldMat`` of an :class:`.Image` overlay.
"""


import logging

import wx

import numpy as np

import fsl.data.image                       as fslimage
import fsl.utils.idle                       as idle
import fsl.transform.affine                 as affine

import fsleyes_props                        as props
import fsleyes_widgets.floatslider          as fslider

import fsleyes.actions                      as actions
import fsleyes.views.orthopanel             as orthopanel
import fsleyes.controls.controlpanel        as ctrlpanel
import fsleyes.displaycontext               as displaycontext
import fsleyes.strings                      as strings
import fsleyes.plugins.tools.applyflirtxfm  as applyflirtxfm
import fsleyes.plugins.tools.saveflirtxfm   as saveflirtxfm
import fsleyes.controls.displayspacewarning as dswarning


log = logging.getLogger(__name__)


class EditTransformAction(actions.ToggleControlPanelAction):
    """The ``EditTransformAction`` just toggles an
    :class:`.EditTransformPanel`. It is added under the FSLeyes Tools menu.
    """

    @staticmethod
    def supportedViews():
        """The ``EditTransformAction`` is restricted for use with
        :class:`.OrthoPanel` views.
        """
        return [orthopanel.OrthoPanel]


    def __init__(self, overlayList, displayCtx, ortho):
        """Create an ``EditTransformAction``.
        """
        super().__init__(overlayList, displayCtx, EditTransformPanel, ortho)
        self.__ortho = ortho

        displayCtx.addListener('selectedOverlay', self.name,
                               self.__selectedOverlayChanged)


    def destroy(self):
        """Called when the :class:`.OrthoPanel` that owns this action is
        closed. Clears references, removes listeners, and calls the base
        class ``destroy`` method.
        """
        if self.destroyed:
            return

        self.__ortho = None
        self.displayCtx.removeListener('selectedOverlay', self.name)
        super().destroy()


    def __selectedOverlayChanged(self, *a):
        """Called when the selected overlay changes. Enables/disables this
        action (and hence the bound Tools menu item) depending on whether the
        overlay is an image.
        """
        ovl = self.displayCtx.getSelectedOverlay()
        self.enabled = isinstance(ovl, fslimage.Image)


class EditTransformPanel(ctrlpanel.ControlPanel):
    """The :class:`EditTransformPanel` class is a FSLeyes control panel which
    allows the user to adjust the ``voxToWorldMat`` of an :class:`.Image`
    overlay.


    Controls are provided allowing the user to construct a transformation
    matrix from scales, offsets, and rotations. While the user is adjusting
    the transformation, the :attr:`.NiftiOpts.displayXform` is used to
    update the overlay display in real time. When the user clicks the *Apply*
    button, the transformation is applied to the image's ``voxToWorldMat``
    attribute.


    This panel also has buttons which allow the user to load/save the
    transformation matrix - they use functions in the :mod:`.applyflirtxfm`
    and :mod:`.saveflirtxfm` modules to load, save, and calculate
    transformation matrices. When the user loads a matrix, it is used in place
    of the :attr:`.Image.voxToWorldMat` transformation.

    .. note:: The effect of editing the transformation will only be visible
              if the :attr:`.DisplayContext.displaySpace` is set to
              ``'world'``, or to some image which is not being edited. A
              warning is shown at the top of the panel if the ``displaySpace``
              is not set appropriately.
    """


    @staticmethod
    def ignoreControl():
        """Tells the FSLeyes plugin system not to add the
        ``EditTransformPanel`` as an option to the FSLeyes settings menu.
        Instead, the :class:`EditTransformAction` action is added to the
        tools menu.
        """
        return True


    @staticmethod
    def supportedViews():
        """Overrides :meth:`.ControlMixin.supportedViews`. The
        ``EditTransformPanel`` is only intended to be added to
        :class:`.OrthoPanel` views.
        """
        return [orthopanel.OrthoPanel]


    @staticmethod
    def defaultLayout():
        """Returns a dictionary of arguments to be passed to the
        :meth:`.ViewPanel.togglePanel` method when an ``EditTransformPanel``
        is created.
        """
        return dict(floatPane=True, floatOnly=True)


    def __init__(self, parent, overlayList, displayCtx, ortho):
        """Create an ``EditTransformPanel``.

        :arg parent:      The :mod:`wx` parent object.
        :arg overlayList: The :class:`.OverlayList` instance.
        :arg displayCtx:  The :class:`.DisplayContext` instance.
        :arg ortho:       The :class:`.OrthoPanel` instance.
        """

        ctrlpanel.ControlPanel.__init__(
            self, parent, overlayList, displayCtx, ortho)

        self.__ortho = ortho

        # A ref to the currently selected
        # (compatible) overlay is kept here.
        # The __extraXform attribute is used
        # to store a FLIRT transform if the
        # user has loaded one. This 'extra'
        # matrix is used in place of the
        # image voxToWorldMat (i.e. its sform);
        # the scale/offset/ rotate transform
        # defined by the widgets on this panel
        # is still applied.
        #
        # In the future, I might allow the
        # user to load/apply an arbitrary
        # (non-FLIRT) transform.
        self.__overlay    = None
        self.__extraXform = None

        # When the selected overlay is changed, the
        # transform settings for the previously selected
        # overlay are cached in this dict, so they can be
        # restored if/when the overlay is re-selected.
        #
        # { overlay : (scales, offsets, rotations, extraXform) }
        self.__cachedXforms = {}

        scArgs = {
            'value'    : 0,
            'minValue' : 0.001,
            'maxValue' : 3,
            'style'    : fslider.SSP_NO_LIMITS
        }

        offArgs = {
            'value'    : 0,
            'minValue' : -250,
            'maxValue' :  250,
            'style'    : fslider.SSP_NO_LIMITS
        }

        rotArgs = {
            'value'    : 0,
            'minValue' : -180,
            'maxValue' :  180,
            'style'    : 0
        }

        # rotate about the centre of the image,
        # or the current world location
        centreOpts   = ['volume', 'cursor']
        centreLabels = [strings.labels[self, 'centre.options'][o]
                        for o in centreOpts]

        self.__overlayName = wx.StaticText(self)
        self.__dsWarning   = dswarning.DisplaySpaceWarning(
            self,
            overlayList,
            displayCtx,
            self.frame,
            strings.labels[self, 'dsWarning'],
            'overlay',
            'world')

        self.__xscale  = fslider.SliderSpinPanel(self, label='X', **scArgs)
        self.__yscale  = fslider.SliderSpinPanel(self, label='Y', **scArgs)
        self.__zscale  = fslider.SliderSpinPanel(self, label='Z', **scArgs)

        self.__xoffset = fslider.SliderSpinPanel(self, label='X', **offArgs)
        self.__yoffset = fslider.SliderSpinPanel(self, label='Y', **offArgs)
        self.__zoffset = fslider.SliderSpinPanel(self, label='Z', **offArgs)

        self.__xrotate = fslider.SliderSpinPanel(self, label='X', **rotArgs)
        self.__yrotate = fslider.SliderSpinPanel(self, label='Y', **rotArgs)
        self.__zrotate = fslider.SliderSpinPanel(self, label='Z', **rotArgs)
        self.__centre  = wx.Choice(self)

        self.__scaleLabel  = wx.StaticText(self)
        self.__offsetLabel = wx.StaticText(self)
        self.__rotateLabel = wx.StaticText(self)
        self.__centreLabel = wx.StaticText(self)

        self.__oldXformLabel = wx.StaticText(self)
        self.__oldXform      = wx.StaticText(self)
        self.__newXformLabel = wx.StaticText(self)
        self.__newXform      = wx.StaticText(self)

        self.__apply     = wx.Button(self)
        self.__reset     = wx.Button(self)
        self.__loadFlirt = wx.Button(self)
        self.__saveFlirt = wx.Button(self)
        self.__cancel    = wx.Button(self)

        self.__overlayName  .SetLabel(strings.labels[self, 'noOverlay'])
        self.__scaleLabel   .SetLabel(strings.labels[self, 'scale'])
        self.__offsetLabel  .SetLabel(strings.labels[self, 'offset'])
        self.__rotateLabel  .SetLabel(strings.labels[self, 'rotate'])
        self.__centreLabel  .SetLabel(strings.labels[self, 'centre'])
        self.__apply        .SetLabel(strings.labels[self, 'apply'])
        self.__reset        .SetLabel(strings.labels[self, 'reset'])
        self.__loadFlirt    .SetLabel(strings.labels[self, 'loadFlirt'])
        self.__saveFlirt    .SetLabel(strings.labels[self, 'saveFlirt'])
        self.__cancel       .SetLabel(strings.labels[self, 'cancel'])
        self.__oldXformLabel.SetLabel(strings.labels[self, 'oldXform'])
        self.__newXformLabel.SetLabel(strings.labels[self, 'newXform'])

        self.__centre.Set(centreLabels)
        self.__centreOpts = centreOpts

        # Populate the xform labels with a
        # dummy xform, so an appropriate
        # minimum size will get calculated
        # below
        self.__formatXform(np.eye(4), self.__oldXform)
        self.__formatXform(np.eye(4), self.__newXform)

        self.__primarySizer   = wx.BoxSizer(wx.VERTICAL)
        self.__secondarySizer = wx.BoxSizer(wx.HORIZONTAL)
        self.__controlSizer   = wx.BoxSizer(wx.VERTICAL)
        self.__xformSizer     = wx.BoxSizer(wx.VERTICAL)
        self.__buttonSizer    = wx.BoxSizer(wx.HORIZONTAL)

        self.__primarySizer  .Add((1, 10),            flag=wx.EXPAND)
        self.__primarySizer  .Add(self.__overlayName, flag=wx.CENTRE)
        self.__primarySizer  .Add(self.__dsWarning,   flag=wx.CENTRE)
        self.__primarySizer  .Add((1, 10),            flag=wx.EXPAND)
        self.__primarySizer  .Add(self.__secondarySizer)
        self.__primarySizer  .Add((1, 10),            flag=wx.EXPAND)
        self.__primarySizer  .Add(self.__buttonSizer, flag=wx.EXPAND)
        self.__primarySizer  .Add((1, 10),            flag=wx.EXPAND)

        self.__secondarySizer.Add((10, 1),           flag=wx.EXPAND)
        self.__secondarySizer.Add(self.__controlSizer)
        self.__secondarySizer.Add((10, 1),           flag=wx.EXPAND)
        self.__secondarySizer.Add(self.__xformSizer, flag=wx.EXPAND)
        self.__secondarySizer.Add((10, 1),           flag=wx.EXPAND)

        self.__controlSizer.Add(self.__scaleLabel)
        self.__controlSizer.Add(self.__xscale)
        self.__controlSizer.Add(self.__yscale)
        self.__controlSizer.Add(self.__zscale)
        self.__controlSizer.Add(self.__offsetLabel)
        self.__controlSizer.Add(self.__xoffset)
        self.__controlSizer.Add(self.__yoffset)
        self.__controlSizer.Add(self.__zoffset)
        self.__controlSizer.Add(self.__rotateLabel)
        self.__controlSizer.Add(self.__xrotate)
        self.__controlSizer.Add(self.__yrotate)
        self.__controlSizer.Add(self.__zrotate)
        self.__controlSizer.Add(self.__centreLabel)
        self.__controlSizer.Add(self.__centre)

        self.__xformSizer.Add((1, 1), flag=wx.EXPAND, proportion=1)
        self.__xformSizer.Add(self.__oldXformLabel)
        self.__xformSizer.Add(self.__oldXform)
        self.__xformSizer.Add((1, 1), flag=wx.EXPAND, proportion=1)
        self.__xformSizer.Add(self.__newXformLabel)
        self.__xformSizer.Add(self.__newXform)
        self.__xformSizer.Add((1, 1), flag=wx.EXPAND, proportion=1)

        self.__buttonSizer.Add((10, 1),          flag=wx.EXPAND, proportion=1)
        self.__buttonSizer.Add(self.__apply,     flag=wx.EXPAND)
        self.__buttonSizer.Add((10, 1),          flag=wx.EXPAND)
        self.__buttonSizer.Add(self.__reset,     flag=wx.EXPAND)
        self.__buttonSizer.Add((10, 1),          flag=wx.EXPAND)
        self.__buttonSizer.Add(self.__loadFlirt, flag=wx.EXPAND)
        self.__buttonSizer.Add((10, 1),          flag=wx.EXPAND)
        self.__buttonSizer.Add(self.__saveFlirt, flag=wx.EXPAND)
        self.__buttonSizer.Add((10, 1),          flag=wx.EXPAND)
        self.__buttonSizer.Add(self.__cancel,    flag=wx.EXPAND)
        self.__buttonSizer.Add((10, 1),          flag=wx.EXPAND, proportion=1)

        self.SetSizer(self.__primarySizer)
        self.SetMinSize(self.__primarySizer.GetMinSize())

        self.__xscale .Bind(fslider.EVT_SSP_VALUE, self.__xformChanged)
        self.__yscale .Bind(fslider.EVT_SSP_VALUE, self.__xformChanged)
        self.__zscale .Bind(fslider.EVT_SSP_VALUE, self.__xformChanged)
        self.__xoffset.Bind(fslider.EVT_SSP_VALUE, self.__xformChanged)
        self.__yoffset.Bind(fslider.EVT_SSP_VALUE, self.__xformChanged)
        self.__zoffset.Bind(fslider.EVT_SSP_VALUE, self.__xformChanged)
        self.__xrotate.Bind(fslider.EVT_SSP_VALUE, self.__xformChanged)
        self.__yrotate.Bind(fslider.EVT_SSP_VALUE, self.__xformChanged)
        self.__zrotate.Bind(fslider.EVT_SSP_VALUE, self.__xformChanged)
        self.__centre .Bind(wx.EVT_CHOICE,         self.__xformChanged)

        self.__apply    .Bind(wx.EVT_BUTTON, self.__onApply)
        self.__reset    .Bind(wx.EVT_BUTTON, self.__onReset)
        self.__loadFlirt.Bind(wx.EVT_BUTTON, self.__onLoadFlirt)
        self.__saveFlirt.Bind(wx.EVT_BUTTON, self.__onSaveFlirt)
        self.__cancel   .Bind(wx.EVT_BUTTON, self.__onCancel)

        displayCtx .addListener('selectedOverlay',
                                self.name,
                                self.__selectedOverlayChanged)
        overlayList.addListener('overlays',
                                self.name,
                                self.__selectedOverlayChanged)

        self.__selectedOverlayChanged()


    def destroy(self):
        """Must be called when this ``EditTransformPanel`` is no longer
        needed. Removes listeners and cleans up references.
        """

        self.__deregisterOverlay()

        displayCtx  = self.displayCtx
        overlayList = self.overlayList
        dsWarning   = self.__dsWarning

        displayCtx .removeListener('selectedOverlay', self.name)
        overlayList.removeListener('overlays',        self.name)

        self.__ortho        = None
        self.__cachedXforms = None
        self.__dsWarning    = None

        dsWarning.destroy()
        ctrlpanel.ControlPanel.destroy(self)


    def __registerOverlay(self, overlay):
        """Called by :meth:`__selectedOverlayChanged`. Stores a reference
        to the given ``overlay``.
        """

        self.__overlay = overlay
        display = self.displayCtx.getDisplay(overlay)
        display.addListener('name', self.name, self.__overlayNameChanged)

        self.__overlayNameChanged()


    def __deregisterOverlay(self):
        """Called by :meth:`__selectedOverlayChanged`. Clears references
        to the most recently registered overlay.
        """

        if self.__overlay is None:
            return

        overlay = self.__overlay

        scales, offsets, rotations, centre = self.__getCurrentXformComponents()
        extra                              = self.__extraXform

        self.__cachedXforms[overlay] = (scales, offsets, rotations,
                                        centre, extra)

        self.__overlay    = None
        self.__extraXform = None

        self.__overlayName.SetLabel(strings.labels[self, 'noOverlay'])

        # Catch errors in case the
        # overlay has been removed
        # from the list
        try:
            display = self.displayCtx.getDisplay(overlay)
            display.removeListener('name', self.name)

        except displaycontext.InvalidOverlayError:
            pass


    def __overlayNameChanged(self, *a):
        """Called when the :attr:`.Display.name` of the currently selected
        overlay changes. Updates the name label.
        """
        display = self.displayCtx.getDisplay(self.__overlay)
        label   = strings.labels[self, 'overlayName'].format(display.name)

        self.__overlayName.SetLabel(label)


    def __selectedOverlayChanged(self, *a):
        """Called when the :attr:`.DisplayContext.selectedOverlay` or
        :attr:`.OverlayList.overlays` properties change. If the newly
        selected overlay is an :class:`.Image`, it is registered, and
        the transform widgets reset.
        """
        overlay = self.displayCtx.getSelectedOverlay()

        if overlay is self.__overlay:
            return

        self.__deregisterOverlay()

        enabled = isinstance(overlay, fslimage.Image)

        self.Enable(enabled)

        if not enabled:
            return

        self.__registerOverlay(overlay)

        xform                                     = overlay.voxToWorldMat
        scales, offsets, rotations, centre, extra = self.__cachedXforms.get(
            overlay, ((1, 1, 1), (0, 0, 0), (0, 0, 0), 'volume', None))

        self.__extraXform = extra

        self.__formatXform(xform, self.__oldXform)

        # TODO Set limits based on image size?
        self.__xscale .SetValue(scales[   0])
        self.__yscale .SetValue(scales[   1])
        self.__zscale .SetValue(scales[   2])
        self.__xoffset.SetValue(offsets[  0])
        self.__yoffset.SetValue(offsets[  1])
        self.__zoffset.SetValue(offsets[  2])
        self.__xrotate.SetValue(rotations[0])
        self.__yrotate.SetValue(rotations[1])
        self.__zrotate.SetValue(rotations[2])
        self.__centre .SetSelection(self.__centreOpts.index(centre))

        self.__xformChanged()


    def __formatXform(self, xform, ctrl):
        """Format the given ``xform``  on the given ``wx.StaticText``
        ``ctrl``.
        """

        text = ''

        for rowi in range(xform.shape[0]):
            for coli in range(xform.shape[1]):

                text = text + '{: 9.2f} '.format(xform[rowi, coli])

            text = text + '\n'

        ctrl.SetLabel(text)


    def __getCurrentXformComponents(self):
        """Returns the components of the transformation matrix defined
        by the scale, offset and rotation widgets.
        """
        scales    = [self.__xscale .GetValue(),
                     self.__yscale .GetValue(),
                     self.__zscale .GetValue()]
        offsets   = [self.__xoffset.GetValue(),
                     self.__yoffset.GetValue(),
                     self.__zoffset.GetValue()]
        rotations = [self.__xrotate.GetValue(),
                     self.__yrotate.GetValue(),
                     self.__zrotate.GetValue()]
        centre    = self.__centreOpts[self.__centre.GetSelection()]

        return scales, offsets, rotations, centre


    def __getCurrentXform(self):
        """Returns the current transformation matrix defined by the scale,
        offset, and rotation widgets.
        """

        scales, offsets, rotations, centre = self.__getCurrentXformComponents()

        rotations = [r * np.pi / 180 for r in rotations]

        if centre == 'volume':
            # We need to figure out the centre
            # of the image in world coordinates
            # to define the origin of rotation.
            shape  = self.__overlay.shape
            lo, hi = affine.axisBounds(shape, self.__overlay.voxToWorldMat)
            origin = [l + (h - l) / 2.0 for h, l in zip(hi, lo)]
        else:
            origin = self.displayCtx.worldLocation

        return affine.compose(scales, offsets, rotations, origin)


    def __xformChanged(self, ev=None):
        """Called when any of the scale, offset, or rotate widgets are
        modified. Updates the :attr:`.NiftiOpts.displayXform` for the
        overlay currently being edited.
        """

        if self.__overlay is None:
            return

        overlay  = self.__overlay
        opts     = self.displayCtx.getOpts(overlay)

        if self.__extraXform is None: v2wXform = overlay.voxToWorldMat
        else:                         v2wXform = self.__extraXform

        xform = self.__getCurrentXform()
        xform = affine.concat(xform, v2wXform)

        self.__formatXform(xform, self.__newXform)

        # The NiftiOpts.displayXform is applied on
        # top of the image voxToWorldMat. But our
        # xform here has been constructed to replace
        # the voxToWorldMat entirely. So we include
        # a worldToVoxMat transform to trick the
        # NiftiOpts code.
        opts.displayXform = affine.concat(xform, overlay.worldToVoxMat)


    def __onApply(self, ev):
        """Called when the *Apply* button is pushed. Sets the
        ``voxToWorldMat`` attribute of the :class:`.Image` instance being
        transformed.
        """

        overlay = self.__overlay

        if overlay is None:
            return

        if self.__extraXform is None: v2wXform = overlay.voxToWorldMat
        else:                         v2wXform = self.__extraXform

        newXform = self.__getCurrentXform()
        opts     = self.displayCtx.getOpts(overlay)

        xform = affine.concat(newXform, v2wXform)

        with props.suppress(opts, 'displayXform'):
            opts.displayXform     = np.eye(4)
            overlay.voxToWorldMat = xform

        # Reset the interface, and clear any
        # cached transform for this overlay
        self.__deregisterOverlay()
        self.__cachedXforms.pop(overlay, None)
        self.__selectedOverlayChanged()


    def __resetAllOverlays(self):
        """Resets the :attr:`.NiftiOpts.displayXform` matrix for
        all overlays that have been modified, and clears the internal
        transformation matrix cache.

        This method is called by :meth:`__onReset` and :meth:`__onCancel`.
        """

        reset = list(self.__cachedXforms.keys())

        if self.__overlay is not None:
            reset.append(self.__overlay)

        self.__deregisterOverlay()
        self.__cachedXforms = {}

        for overlay in reset:
            try:
                opts = self.displayCtx.getOpts(overlay)
                opts.displayXform = np.eye(4)

            # In cas overlay has been removed
            except displaycontext.InvalidOverlayError:
                pass


    def __onReset(self, ev=None):
        """Called when the *Reset* button is pushed. Resets the
        transformation.
        """

        self.__resetAllOverlays()
        self.__selectedOverlayChanged()


    def __onLoadFlirt(self, ev):
        """Called when the user clicks the *Load FLIRT transform* button.
        Prompts the user to choose a FLIRT transformation matrix and reference
        image, and then applies the transformation.
        """

        overlay = self.__overlay

        if overlay is None:
            return

        overlayList               = self.overlayList
        displayCtx                = self.displayCtx
        affType, matFile, refFile = applyflirtxfm.promptForFlirtFiles(
            self,
            overlay,
            overlayList,
            displayCtx)

        if all((affType is None, matFile is None, refFile is None)):
            return

        if affType == 'flirt':
            xform = applyflirtxfm.calculateTransform(
                overlay,
                overlayList,
                displayCtx,
                matFile,
                refFile)

        elif affType == 'v2w':
            xform = np.loadtxt(matFile)

        self.__extraXform = xform
        self.__xformChanged()


    def __onSaveFlirt(self, ev):
        """Called when the user clicks the *Save FLIRT* button. Saves the
        current transformation to a FLIRT matrix file.
        """

        overlay = self.__overlay

        if overlay is None:
            return

        overlayList               = self.overlayList
        displayCtx                = self.displayCtx
        affType, matFile, refFile = applyflirtxfm.promptForFlirtFiles(
            self,
            overlay,
            overlayList,
            displayCtx,
            save=True)

        if all((affType is None, matFile is None, refFile is None)):
            return

        if self.__extraXform is None: v2wXform = overlay.voxToWorldMat
        else:                         v2wXform = self.__extraXform

        newXform = self.__getCurrentXform()
        v2wXform = affine.concat(newXform, v2wXform)

        if affType == 'flirt':
            xform = saveflirtxfm.calculateTransform(
                overlay,
                overlayList,
                displayCtx,
                refFile,
                srcXform=v2wXform)
        elif affType == 'v2w':
            xform = v2wXform

        try:
            np.savetxt(matFile, xform, fmt='%0.10f')

        except Exception as e:

            log.warning('Error saving FLIRT matrix: %s', e)

            wx.MessageDialog(
                self,
                strings.messages[self, 'saveFlirt.error'].format(str(e)),
                style=wx.ICON_ERROR).ShowModal()


    def __onCancel(self, ev=None):
        """Called when the *Cancel* button is pushed. Resets the
        :attr:`.NiftiOpts.displayXform` attribute of the overlay being
        transformed, and then calls
        :meth:`.OrthoPanel.toggleEditTransformPanel` to close this panel.
        """

        self.__resetAllOverlays()
        idle.idle(self.__ortho.togglePanel, type(self))



#!/usr/bin/env python
#
# glsh.py - The GLSH class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`GLSH` class, a :class:`.GLObject` for
rendering :class:`.Image` overlays which contain spherical harmonic (SH)
coefficients which represent fibre orientation distributions (FODs).  The
``GLSH`` class uses functions defined in the :mod:`.gl21.glsh_funcs` module.

:class:`GLSH` instances can only be rendered in OpenGL 2.1 and above.
"""


import               logging
import               warnings

import numpy      as np

import OpenGL.GL  as gl


import fsleyes.gl          as fslgl
import fsleyes.gl.textures as textures
import fsleyes.gl.glvector as glvector


log = logging.getLogger(__name__)


class GLSH(glvector.GLVectorBase):
    """The ``GLSH`` class is a :class:`.GLVectorBase` for rendering
    :class:`.Image` overlays which contain spherical harmonic (SH) coefficients
    that represent fibre orientation distributions (FOD).


    This class manages listeners on :class:`.Display` and :class:`.SHOpts`
    instances, and also manages two textures (described below). The rendering
    logic is implemented in the :mod:`.gl21.glsh_funcs` module.


    Each voxel in a FOD image contains coefficients which describe a linear
    combination of spherical harmonic functions.  The
    :meth:`SHOpts.getSHParameters` method returns a numpy array which may be
    used to transform these coefficients into a set of radii that can be
    applied to the vertices of a sphere to visualise the spherical
    function. Pre-calculated vertices of tessellated spheres are used, and are
    retrieved via the :meth:`.SHOpts.getVertices` and
    :meth:`.SHOpts.getIndices` methods.


    These radii are calculated on every call to :meth:`draw` (via the
    :meth:`updateRadTexture` method), and stored in a :class:`.Texture3D`
    instance, which is available as an attribute called ``radTexture``. This
    texture is only 3D out of necessity - it is ultimately interpreted by
    the ``glsh`` vertex shader as a 1D sequence of values, ordered by voxel
    then vertex.


    The radius texture managed by a ``GLSH`` instance is bound to GL
    texture unit ``GL_TEXTURE4``.


    The following attributes are available on a ``GLSH`` instance (and are
    assumed to be present by the functions in the :mod:`.glsh_funcs` module`):


    ============== =====================================================
    ``radTexture`` :class:`.Texture3D` containing radius values for each
                    vertex to be displayed in the current draw call.

    ``vertices``   ``numpy`` array of shape ``(N, 3)`` which comprise a
                   tessellated sphere. The vertex shader will apply
                   the radii to the vertices contained in this array,
                   to form the FODs at every voxel.

    ``indices``    Indices into ``vertices`` defining the faces of the
                   sphere.

    ``nVertices``  Total number of rendered vertices (equal to
                   ``len(indices)``).

    ``vertIdxs``   Indices for each vertex (equal to
                   ``np.arange(vertices.shape[0])``).
    ============== =====================================================
    """


    def __init__(self, image, overlayList, displayCtx, canvas, threedee):
        """Create a ``GLSH`` object.


        Creates a :class:`.Texture3D` instance to store vertex radii, adds
        property listeners to the :class:`.Display` and :class:`.SHOpts`
        instances, and sets up shaders.


        :arg image:       The :class:`.Image` instance
        :arg overlayList: The :class:`.OverlayList`
        :arg displayCtx:  The :class:`.DisplayContext` managing the scene.
        :arg canvas:      The canvas doing the drawing.
        :arg threedee:    Set up for 2D or 3D rendering.
        """

        self.shader     = None
        self.radTexture = None

        # These are updated in the
        # __shStateChanged method.
        self.__shParams = None

        # This texture gets updated on
        # draw calls, so we want it to
        # run on the main thread.
        self.radTexture = textures.Texture3D('{}_radTexture'.format(id(self)),
                                             threaded=False)

        # Usin preinit here (see GLVectorBase)
        # because the GLObject init has to be
        # called before compileShaders can be
        # called, and shStateChanged has to be
        # called after everything else is called.
        glvector.GLVectorBase.__init__(
            self,
            image,
            overlayList,
            displayCtx,
            canvas,
            threedee,
            preinit=self.compileShaders,
            init=self.__shStateChanged)


    def destroy(self):
        """Removes property listeners, destroys textures, and calls
        :func:`.glsh_funcs.destroy`.
        """

        self.removeListeners()

        fslgl.glsh_funcs.destroy(self)

        if self.radTexture is not None:
            self.radTexture.destroy()

        self.radTexture = None

        glvector.GLVectorBase.destroy(self)


    def ready(self):
        """Overrides :class:`.GLVectorBase.ready`. Returns ``True`` when this
        ``GLSH`` is ready to be drawn.
        """
        return self.__shParams is not None \
            and glvector.GLVectorBase.ready(self)


    def getDataResolution(self, xax, yax):
        """Overrides :meth:`.GLVector.getDataResolution`. Returns a pixel
        resolution suitable for off-screen rendering of this ``GLSH``.
        """

        res       = list(glvector.GLVector.getDataResolution(self, xax, yax))
        res[xax] *= 20
        res[yax] *= 20

        return res


    def addListeners(self):
        """Overrides :meth:`.GLVectorBase.addListeners`.

        Called by :meth:`__init__`. Adds listeners to properties of the
        :class:`.Display` and :class:`.SHOpts` instanecs associated with the
        image.
        """

        glvector.GLVectorBase.addListeners(self)

        opts = self.opts
        name = self.name

        opts.addListener('shResolution' ,   name, self.__shStateChanged,
                         immediate=True)
        opts.addListener('shOrder'      ,   name, self.__shStateChanged,
                         immediate=True)
        opts.addListener('size',            name, self.updateShaderState)
        opts.addListener('lighting',        name, self.updateShaderState)
        opts.addListener('orientFlip',      name, self.updateShaderState)
        opts.addListener('colourMode',      name, self.updateShaderState)
        opts.addListener('radiusThreshold', name, self.notify)
        opts.addListener('normalise',       name, self.notify)


    def removeListeners(self):
        """Overrides :meth:`.GLVectorBase.removeListeners`. Called by
        :meth:`destroy`. Removes listeners added by :meth:`addListeners`.
        """

        glvector.GLVectorBase.removeListeners(self)

        opts = self.opts
        name = self.name

        opts.removeListener('shResolution',    name)
        opts.removeListener('shOrder',         name)
        opts.removeListener('size',            name)
        opts.removeListener('lighting',        name)
        opts.removeListener('orientFlip',      name)
        opts.removeListener('colourMode',      name)
        opts.removeListener('radiusThreshold', name)
        opts.removeListener('normalise',       name)


    def compileShaders(self, *a):
        """Overrides :meth:`.GLVectorBase.compileShaders`. Calls
        :func:`.glsh_funcs.compileShaders`.
        """
        fslgl.glsh_funcs.compileShaders(self)


    def updateShaderState(self, *a, **kwa):
        """Overrides :meth:`.GLVectorBase.updateShaderState`. Calls
        :func:`.glsh_funcs.updateShaderState`.
        """
        alwaysNotify = kwa.pop('alwaysNotify', False)

        if fslgl.glsh_funcs.updateShaderState(self) or alwaysNotify:
            self.notify()
            return True
        return False


    def __shStateChanged(self, *a):
        """Called when the :attr:`.SHOpts.shResolution` property changes.
        Re-loads the SH parameters from disk, and attaches them as an
        attribute called ``__shParams``.
        """

        opts = self.opts

        self.__shParams = opts.getSHParameters()
        self.vertices   = opts.getVertices()
        self.indices    = opts.getIndices()
        self.nVertices  = len(self.indices)
        self.vertIdxs   = np.arange(self.vertices.shape[0], dtype=np.float32)

        self.updateShaderState(alwaysNotify=True)


    def __coefVolumeMask(self):
        """Figures out which volumes from the image need to be included in the
        SH radius calculation. If an image has been generated with a particular
        maximum SH function order, but is being displayed at a reduced order,
        a sub-set of the volumes need to be used in the calculation.


        :returns A ``slice`` object which can be used to select a subset of
                 volumes from the SH image.


        For a symmetric SH image (which only contains SH functions of even
        order), each volume corresponds to

        ======  =============  =====
        Volume  Maximum order  Order
        ------  -------------  -----
        0       0               0
        1       2              -2
        2       2              -1
        3       2               0
        4       2               1
        5       2               2
        6       4              -4
        7       4              -3
        8       4              -2
        9       4              -1
        10      4               0
        11      4               1
        12      4               2
        13      4               3
        14      4               4
        15      6              -6
        ...     ...            ...
        ======  =============  =====


        Asymmetric images (containing SH functions of both even and odd
        order) follow the same pattern:


        ======  =============  =====
        Volume  Maximum order  Order
        ------  -------------  -----
        0       0               0
        1       1              -1
        2       1               0
        3       1               1
        4       2              -2
        5       2              -1
        6       2               0
        7       2               1
        8       2               2
        9       3              -3
        10      3              -2
        11      3              -1
        12      3               0
        13      3               1
        14      3               2
        15      3               3
        16      4              -4
        ...     ...            ...
        ======  =============  =====
        """
        opts      = self.opts
        maxOrder  = opts.maxOrder
        dispOrder = opts.shOrder
        shType    = opts.shType
        nvols     = self.image.shape[3]

        if maxOrder == dispOrder:
            return slice(None)

        if shType == 'sym':
            for o in range(dispOrder + 2, maxOrder + 2, 2):
                nvols -= 2 * o + 1

        elif shType == 'asym':
            for o in range(dispOrder + 1, maxOrder + 1):
                nvols -= 2 * o + 1

        return slice(nvols)


    def updateRadTexture(self, voxels):
        """Called by :func:`.glsh_funcs.draw`. Updates the radius texture to
        contain radii for the given set of voxels (assumed to be an ``(N, 3)``
        numpy array).

        If :attr:`.SHOpts.radiusThreshold` is greater than 0, any voxels for
        which all radii are less than the threshold are removed from the
        ``voxels`` array.

        If :attr:`.SHOpts.normalise` is ``True``, the radii within each voxel
        are normalised to lie between 0 and 0.5, so that they fit within the
        voxel.

        This function returns a tuple containing:

          - The ``voxels`` array. If ``SHOpts.radiusThreshold == 0``,
            this will be the same as the input. Otherwise, this will
            be a new array with sub-threshold voxels removed.
            If no voxels are to be rendered (all out of bounds, or below
            the radius threshold), this will be an empty list.

          - The adjusted shape of the radius texture.
        """

        opts = self.opts

        # Remove out-of-bounds voxels
        shape   = self.image.shape[:3]
        x, y, z = voxels.T
        out     = (x <  0)        | \
                  (y <  0)        | \
                  (z <  0)        | \
                  (x >= shape[0]) | \
                  (y >= shape[1]) | \
                  (z >= shape[2])
        voxels  = np.asarray(voxels[~out, :], dtype=np.uint32)
        x, y, z = voxels.T

        # The dot product of the SH parameters with
        # the SH coefficients for a single voxel gives
        # us the radii for every vertex on the FOD
        # sphere. We can calculate the radii for every
        # voxel quickly with a matrix multiplication of
        # the SH parameters with the SH coefficients of
        # *all* voxels.
        params = self.__shParams
        vols   = self.__coefVolumeMask()
        coefs  = self.image.data[x, y, z, vols]
        radii  = np.dot(coefs, params.T)

        # Remove sub-threshold voxels/radii
        if opts.radiusThreshold > 0:
            aboveThres = np.any(radii >= opts.radiusThreshold, axis=1)
            radii      = radii[ aboveThres, :]
            voxels     = voxels[aboveThres, :]

        # No voxels - nothing to do
        if voxels.shape[0] == 0:
            return [], [0, 0, 0]

        # Normalise within voxel if necessary
        if opts.normalise:
            rmin = radii.min(axis=1)
            rmax = radii.max(axis=1)
            with warnings.catch_warnings():
                warnings.filterwarnings('ignore')
                radii = ((radii.T - rmin) / (2 * (rmax - rmin))).T

        # The radii are interpreted as a 1D vector
        # containing the radii for every vertex
        # in every voxel. Because we're using an
        # old version of OpenGL (2.1), we can't
        # simply make this 1D vector available
        # to the vertex shader - we need to copy
        # it into a texture, and that texture has
        # to be 3D, because texture dimensions
        # have a maximum size limit.
        #
        # So here we are calculating a suitable
        # 3D shape in which the radius values can
        # be stored. The shape may end up being
        # larger than necessary, if the number of
        # voxels cannot easily be divided/dispersed
        # across the other dimensions.
        radTexShape = np.array([radii.size, 1, 1])
        maxTexSize  = gl.glGetIntegerv(gl.GL_MAX_3D_TEXTURE_SIZE)
        while np.any(radTexShape > maxTexSize):

            # Find the biggest and smallest dimensions
            imin = np.argmin(radTexShape)
            imax = np.argmax(radTexShape)

            # Try to find a way to move values
            # from the biggest dimension to the
            # smallest dimension
            divisor = 0
            for i in (2, 3, 5, 7):
                if radTexShape[imax] % i == 0:
                    divisor = i
                    break

            # If we can't evenly reshape the texture
            # dimensions, we have to increase the
            # texture size - the radius data will
            # only take up a portion of the allocated
            # texture size.
            else:
                divisor            = 2
                radTexShape[imax] += 1

            radTexShape[imax] /= divisor
            radTexShape[imin] *= divisor

        # Resize and reshape the radius array as needed.
        # Radii starts off as a (voxels, vertices) array.
        # We flatten it to 1D (by voxels, i.e. with
        # vertices the fastest changing)
        radii      = radii.reshape(-1)
        radTexSize = np.prod(radTexShape)

        # if the texture size needs to be bigger,
        # create a new array, and copy the radii
        # data into it
        if radTexSize != radii.size:
            tmp = np.zeros(radTexSize, dtype=radii.dtype)
            tmp[:radii.size] = radii
            radii = tmp

        # reshape to the texture size, making
        # sure that the first dimension is the
        # fastest changing (as this is what
        # OpenGL requires). This ensures that
        # the data remains contiguous in memory
        radii = radii.reshape(radTexShape, order='F')

        # Copy the data to the texture
        self.radTexture.set(data=radii)

        return voxels, radTexShape


    def texturesReady(self):
        """Overrides :meth:`.GLVectorBase.texturesReady`. Returns ``True`` if
        all textures used by this ``GLSH`` instance are ready to be used,
        ``False`` otherwise.
        """
        return (self.radTexture is not None and
                glvector.GLVectorBase.texturesReady(self))


    def preDraw(self, xform=None, bbox=None):
        """Overrides :meth:`.GLVectorBase.preDraw`.  Binds textures, and calls
        :func:`.glsh_funcs.preDraw`.
        """

        # The radTexture needs to be bound *last*,
        # because the updateRadTexture method will
        # be called through draw, and if radTexture
        # is not the most recently bound texture,
        # the update will fail.
        glvector.GLVectorBase.preDraw(self, xform, bbox)
        self.radTexture.bindTexture(gl.GL_TEXTURE4)
        fslgl.glsh_funcs.preDraw(self, xform, bbox)


    def draw2D(self, *args, **kwargs):
        """Overrides :meth:`.GLObject.draw`. Calls :func:`.glsh_funcs.draw`.
        """
        fslgl.glsh_funcs.draw2D(self, *args, **kwargs)


    def draw3D(self, *args, **kwargs):
        """Overrides :meth:`.GLObject.draw`. Calls :func:`.glsh_funcs.draw`.
        """
        fslgl.glsh_funcs.draw3D(self, *args, **kwargs)


    def postDraw(self, xform=None, bbox=None):
        """Overrides :meth:`.GLVectorBase.postDraw`.  Unbinds textures, and
        calls :func:`.glsh_funcs.postDraw`.
        """
        glvector.GLVectorBase.postDraw(self, xform, bbox)
        self.radTexture.unbindTexture()
        fslgl.glsh_funcs.postDraw(self, xform, bbox)



#!/usr/bin/env python
#
# glmesh.py - The GLMesh class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`GLMesh` class, a :class:`.GLObject` used
to render :class:`.Mesh` overlays.
"""


import numpy        as np
import numpy.linalg as npla
import OpenGL.GL    as gl

from . import                  globject
import fsl.data.utils       as dutils
import fsl.transform.affine as affine
import fsleyes.gl           as fslgl
import fsleyes.gl.routines  as glroutines
import fsleyes.gl.textures  as textures


class GLMesh(globject.GLObject):
    """The ``GLMesh`` class is a :class:`.GLObject` which encapsulates the
    logic required to draw 2D slices, and 3D renderings, of a
    :class:`.Mesh` overlay.  The ``GLMesh`` class assumes that the
    :class:`.Display` instance associated with the ``Mesh`` overlay
    holds a reference to a :class:`.MeshOpts` instance, which contains
    ``GLMesh`` specific display settings.


    **2D rendering**


    A ``GLMesh`` is rendered in one of two different ways, depending upon the
    value of the :attr:`.MeshOpts.outline` and :attr:`.MeshOpts.vertexData`
    properties.


    *Cross-sections*

    If ``outline is False and vertexData is None``, a filled cross-section of
    the mesh is drawn. This is accomplished using a three-pass technique,
    roughly following that described at
    http://glbook.gamedev.net/GLBOOK/glbook.gamedev.net/moglgp/advclip.html:

     1. The front face of the mesh is rendered to the stencil buffer.

     2. The back face is rendered to the stencil buffer, and subtracted
        from the front face.

     3. This intersection (of the mesh with a plane at the slice Z location)
        is then rendered to the canvas.


    This cross-section is filled with the :attr:`.MeshOpts.colour` property.


    *Outlines*


    If ``outline is True or vertexData is not None``, the intersection of the
    mesh triangles with the viewing plane is calculated. These lines are then
    rendered as ``GL_LINES`` primitives.


    When a mesh outline is drawn on a canvas, the calculated line vertices,
    and corresponding mesh faces (triangles) are cached via the
    :meth:`.OverlayList.setData` method. This is so that other parts of
    FSLeyes can access this information if necessary (e.g. the
    :class:`.OrthoViewProfile` provides mesh-specific interaction). For a
    canvas which is showing a plane orthogonal to display axis X, Y or Z, this
    data will be given a key of ``'crosssection_0'``, ``'crosssection_1'``, or
    ``'crosssection_2'`` respectively.


    *Colouring*


    These lines will be coloured in one of the following ways:

       - With the ``MeshOpts.colour``.

       - According to the :attr:`.MeshOpts.vertexData` if it is set, in which
         case the properties of the :class:`.ColourMapOpts` class (from which
         the :class:`.MeshOpts` class derives) come into effect.

       - Or, if :attr:`vertexData` is set, and the :attr:`.MeshOpts.useLut`
         property is ``True``, the :attr:`.MeshOpts.lut` is used.


    When ``MeshOpts.vertexData is not None``, the ``GLMesh`` class makes use
    of the ``glmesh`` vertex and fragment shaders. These shaders are managed
    by two OpenGL version-specific modules - :mod:`.gl14.glmesh_funcs`, and
    :mod:`.gl21.glmesh_funcs`. These version specific modules must provide the
    following functions:


     - ``compileShaders(GLMesh)``: Compiles vertex/fragment shaders.

     - ``updateShaderState(GLMesh, **kwargs)``: Updates vertex/fragment
       shaders. Should expect the following keyword arguments:

        - ``useNegCmap``: Boolean which is ``True`` if a negative colour map
           should be used

        - ``cmapXform``:   Transformation matrix which transforms from vertex
                           data into colour map texture coordinates.

        - ``flatColour``:  Colour to use for fragments which are outside
                           of the clipping range.

     - ``preDraw(GLMesh)``: Prepare for drawing (e.g. load shaders)

     - ``draw(GLMesh, glType, vertices, indices=None, normals=None, vdata=None)``:  # noqa
        Draws mesh using shaders.

     - ``postDraw(GLMesh)``: Clean up after drawing


    **3D rendering**


    3D mesh rendering is much simpler than 2D rendering. The mesh is simply
    rendered to the canvas, and coloured in the same way as described above.
    Whether the 3D mesh is coloured with a flat colour, or according to vertex
    data, shader programs are used which colour the mesh, and also apply a
    simple lighting effect.


    **Flat shading**

    When the :attr:`.MeshOpts.flatShading` property is active (only possible
    when 3D rendering), the mesh vertices and indices need to be modified.
    In order to achive flat shading, each vertex of a triangle must be given
    the same vertex data value. This means that vertices cannot be shared by
    different triangles. So when the ``flatShading`` property changes, the
    :meth:`updateVertices` method will re-generate the vertices and indices
    accordingly. The :meth:`getVertexData` method will modify the vertex
    data that gets passed to the shader accordingly too.
    """


    def __init__(self, overlay, overlayList, displayCtx, canvas, threedee):
        """Create a ``GLMesh``.

        :arg overlay:     A :class:`.Mesh` overlay.
        :arg overlayList: The :class:`.OverlayList`
        :arg displayCtx:  The :class:`.DisplayContext` managing the scene.
        :arg canvas:      The canvas drawing this ``GLMesh``.
        :arg threedee:    2D or 3D rendering.
        """

        globject.GLObject.__init__(
            self, overlay, overlayList, displayCtx, canvas, threedee)

        self.flatShader   = None
        self.dataShader   = None
        self.activeShader = None

        # We use a render texture when
        # rendering model cross sections.
        # This texture is kept at display/
        # screen resolution
        self.renderTexture = textures.RenderTexture(
            self.name, interp=gl.GL_NEAREST)

        # Mesh overlays are coloured:
        #
        #  - with a constant colour (opts.outline == False), or
        #
        #  - with a +/- colour map, (opts.vertexData not None), or
        #
        #  - with a lookup table (opts.useLut == True and
        #    opts.vertexData is not None)
        self.cmapTexture    = textures.ColourMapTexture(  self.name)
        self.negCmapTexture = textures.ColourMapTexture(  self.name)
        self.lutTexture     = textures.LookupTableTexture(self.name)

        self.lut = None

        self.registerLut()
        self.addListeners()
        self.updateVertices()
        self.refreshCmapTextures(notify=False)

        self.compileShaders()
        self.updateShaderState()


    def destroy(self):
        """Must be called when this ``GLMesh`` is no longer needed. Removes
        some property listeners and destroys the colour map textures and
        off-screen :class:`.RenderTexture`.
        """

        self.renderTexture .destroy()
        self.cmapTexture   .destroy()
        self.negCmapTexture.destroy()
        self.lutTexture    .destroy()

        self.removeListeners()
        self.deregisterLut()

        globject.GLObject.destroy(self)

        if self.flatShader is not None: self.flatShader.destroy()
        if self.dataShader is not None: self.dataShader.destroy()

        self.dataShader   = None
        self.flatShader   = None
        self.activeShader = None

        self.lut            = None
        self.renderTexture  = None
        self.cmapTexture    = None
        self.negCmapTexture = None
        self.lutTexture     = None


    def ready(self):
        """Overrides :meth:`.GLObject.ready`. Always returns ``True``. """
        return True


    def addListeners(self):
        """Called by :meth:`__init__`. Adds some property listeners to the
        :class:`.Display` and :class:`.MeshOpts` instances so the OpenGL
        representation can be updated when the display properties are changed..
        """

        name    = self.name
        display = self.display
        canvas  = self.canvas
        opts    = self.opts

        def shader(*a):
            self.updateShaderState()
            self.notify()

        def vertices(*a):
            self.updateVertices()
            self.updateShaderState()
            self.notify()

        def refreshCmap(*a):
            self.refreshCmapTextures(notify=False)
            self.updateShaderState()
            self.notify()

        def registerLut(*a):
            self.deregisterLut()
            self.registerLut()
            self.refreshCmapTextures()

        def refresh(*a):
            self.notify()

        self.overlay.register(name, vertices, 'vertices')
        opts   .addListener('bounds',           name, vertices,    weak=False)
        opts   .addListener('colour',           name, shader,      weak=False)
        opts   .addListener('outline',          name, refresh,     weak=False)
        opts   .addListener('outlineWidth',     name, refresh,     weak=False)
        opts   .addListener('wireframe',        name, refresh,     weak=False)
        opts   .addListener('vertexData',       name, shader,      weak=False)
        opts   .addListener('modulateData',     name, shader,      weak=False)
        opts   .addListener('vertexDataIndex',  name, shader,      weak=False)
        opts   .addListener('clippingRange',    name, shader,      weak=False)
        opts   .addListener('modulateRange',    name, shader,      weak=False)
        opts   .addListener('invertClipping',   name, shader,      weak=False)
        opts   .addListener('discardClipped',   name, shader,      weak=False)
        opts   .addListener('cmap',             name, refreshCmap, weak=False)
        opts   .addListener('useNegativeCmap',  name, refreshCmap, weak=False)
        opts   .addListener('negativeCmap',     name, refreshCmap, weak=False)
        opts   .addListener('cmapResolution',   name, refreshCmap, weak=False)
        opts   .addListener('interpolateCmaps', name, refreshCmap, weak=False)
        opts   .addListener('invert',           name, refreshCmap, weak=False)
        opts   .addListener('gamma',            name, refreshCmap, weak=False)
        opts   .addListener('displayRange',     name, refreshCmap, weak=False)
        opts   .addListener('useLut',           name, shader,      weak=False)
        opts   .addListener('lut',              name, registerLut, weak=False)
        opts   .addListener('modulateAlpha',    name, shader,      weak=False)
        opts   .addListener('flatShading',      name, vertices,    weak=False)
        display.addListener('alpha',            name, refreshCmap, weak=False)

        # We don't need to listen for
        # brightness or contrast, because
        # they are linked to displayRange.


    def removeListeners(self):
        """Called by :meth:`destroy`. Removes all of the listeners added by
        the :meth:`addListeners` method.
        """
        self.overlay.deregister(self.name, 'vertices')
        self.opts   .removeListener('bounds',           self.name)
        self.opts   .removeListener('colour',           self.name)
        self.opts   .removeListener('outline',          self.name)
        self.opts   .removeListener('outlineWidth',     self.name)
        self.opts   .removeListener('wireframe',        self.name)
        self.opts   .removeListener('vertexData',       self.name)
        self.opts   .removeListener('modulateData',     self.name)
        self.opts   .removeListener('vertexDataIndex',  self.name)
        self.opts   .removeListener('clippingRange',    self.name)
        self.opts   .removeListener('modulateRange',    self.name)
        self.opts   .removeListener('invertClipping',   self.name)
        self.opts   .removeListener('discardClipped',   self.name)
        self.opts   .removeListener('cmap',             self.name)
        self.opts   .removeListener('useNegativeCmap',  self.name)
        self.opts   .removeListener('negativeCmap',     self.name)
        self.opts   .removeListener('cmapResolution',   self.name)
        self.opts   .removeListener('interpolateCmaps', self.name)
        self.opts   .removeListener('invert',           self.name)
        self.opts   .removeListener('gamma',            self.name)
        self.opts   .removeListener('displayRange',     self.name)
        self.opts   .removeListener('useLut',           self.name)
        self.opts   .removeListener('lut',              self.name)
        self.opts   .removeListener('modulateAlpha',    self.name)
        self.opts   .removeListener('flatShading',      self.name)
        self.display.removeListener('alpha',            self.name)


    def registerLut(self):
        """Registers property listeners with the currently registered
        :class:`.LookupTable` (the :attr:`.MeshOpts.lut` property).
        """

        self.lut = self.opts.lut

        if self.lut is not None:
            for topic in ['label', 'added', 'removed']:
                self.lut.register(self.name, self.refreshCmapTextures, topic)


    def deregisterLut(self):
        """De-registers property listeners from the currently registered
        :class:`.LookupTable`.
        """

        if self.lut is not None:
            for topic in ['label', 'added', 'removed']:
                self.lut.deregister(self.name, topic)

        self.lut = None


    def updateVertices(self, *a):
        """Called by :meth:`__init__`, and when certain display properties
        change. (Re-)generates the mesh vertices, indices and normals (if
        being displayed in 3D). They are stored as attributes called
        ``vertices``, ``indices``, and ``normals`` respectively.
        """

        overlay  = self.overlay
        opts     = self.opts
        threedee = self.threedee
        vertices = overlay.vertices
        indices  = overlay.indices
        normals  = self.overlay.vnormals
        vdata    = opts.getVertexData('vertex')
        xform    = opts.getTransform('mesh', 'display')

        if not np.all(np.isclose(xform, np.eye(4))):
            vertices = affine.transform(vertices, xform)

            if self.threedee:
                nmat    = affine.invert(xform).T
                normals = affine.transform(normals, nmat, vector=True)

        self.origIndices = indices
        indices          = np.asarray(indices.flatten(), dtype=np.uint32)

        # If flatShading is active, we cannot share
        # vertices between triangles, so we generate
        # a set of unique vertices for each triangle,
        # and then re-generate the triangle indices.
        # The original indices are saved above, as
        # they will be used by the getVertexData
        # method to duplicate the vertex data.
        if threedee and (vdata is not None) and opts.flatShading:
            self.vertices = vertices[indices].astype(np.float32)
            self.indices  = np.arange(0, len(self.vertices), dtype=np.uint32)
            normals       = normals[indices, :]
        else:
            self.vertices = np.asarray(vertices, dtype=np.float32)
            self.indices  = indices

        self.vertices = dutils.makeWriteable(self.vertices)
        self.indices  = dutils.makeWriteable(self.indices)

        if self.threedee:
            self.normals = np.array(normals, dtype=np.float32)


    def frontFace(self):
        """Returns the face of the mesh triangles which which will be facing
        outwards, either ``GL_CCW`` or ``GL_CW``. This will differ depending
        on the mesh-to-display transformation matrix.

        This method is only used in 3D rendering.
        """

        if not self.threedee:
            return gl.GL_CCW

        # Only looking at the mesh -> display
        # transform, thus we are assuming that
        # the MVP matrix does not have any
        # negative scales.
        xform = self.opts.getTransform('mesh', 'display')

        if npla.det(xform) > 0: return gl.GL_CCW
        else:                   return gl.GL_CW


    def getDisplayBounds(self):
        """Overrides :meth:`.GLObject.getDisplayBounds`. Returns a bounding
        box which contains the mesh vertices.
        """
        return (self.opts.bounds.getLo(),
                self.opts.bounds.getHi())


    def draw2DOutlineEnabled(self):
        """Only relevent for 2D rendering. Returns ``True`` if outline mode
        should be used, ``False`` otherwise.
        """

        opts    = self.opts
        overlay = self.overlay

        return ((overlay.trimesh is not None) and
                (opts.outline or opts.vertexData is not None))


    def needShader(self):
        """Returns ``True`` if a shader should be loaded, ``False`` otherwise.
        Relevant for both 2D and 3D rendering.
        """
        return (self.threedee or
                (self.draw2DOutlineEnabled() and
                 self.opts.vertexData is not None))


    def preDraw(self, xform=None, bbox=None):
        """Overrides :meth:`.GLObject.preDraw`. Performs some pre-drawing
        configuration, which might involve loading shaders, and/or setting the
        size of the backing :class:`.RenderTexture` instance based on the
        current viewport size.
        """

        useShader  = self.needShader()
        outline    = self.draw2DOutlineEnabled()
        useTexture = not (self.threedee or outline)

        # A shader program is used either in 3D, or
        # in 2D when some vertex data is being shown
        if useShader:
            fslgl.glmesh_funcs.preDraw(self)

            if self.opts.vertexData is not None:
                if self.opts.useLut:
                    self.lutTexture.bindTexture(gl.GL_TEXTURE0)
                else:
                    self.cmapTexture   .bindTexture(gl.GL_TEXTURE0)
                    self.negCmapTexture.bindTexture(gl.GL_TEXTURE1)

        # An off-screen texture is used when
        # drawing off-screen textures in 2D
        if useTexture:
            size   = gl.glGetIntegerv(gl.GL_VIEWPORT)
            width  = size[2]
            height = size[3]

            # We only need to resize the texture when
            # the viewport size/quality changes.
            if self.renderTexture.shape != (width, height):
                self.renderTexture.shape = width, height


    def draw2D(self, zpos, axes, xform=None, bbox=None):
        """Overrids :meth:`.GLObject.draw2D`. Draws a 2D slice of the
        :class:`.Mesh`, at the specified Z location.
        """

        overlay       = self.overlay
        lo,  hi       = self.getDisplayBounds()
        xax, yax, zax = axes
        outline       = self.draw2DOutlineEnabled()

        # Mesh is 2D, and is
        # perpendicular to
        # the viewing plane
        if np.any(np.isclose([lo[xax], lo[yax]], [hi[xax], hi[yax]])):
            return

        is2D = np.isclose(lo[zax], hi[zax])

        # 2D meshes are always drawn,
        # regardless of the zpos
        if not is2D and (zpos < lo[zax] or zpos > hi[zax]):
            return

        # the calculateIntersection method caches
        # cross section vertices - make sure they're
        # cleared in case we are not doing an outline
        # draw.
        self.overlayList.setData(overlay, 'crosssection_{}'.format(zax), None)

        if is2D:
            self.draw2DMesh(xform, bbox)

        elif outline:
            self.drawOutline(zpos, axes, xform, bbox)

        else:
            lo, hi = self.calculateViewport(lo, hi, axes, bbox)
            xmin   = lo[xax]
            xmax   = hi[xax]
            ymin   = lo[yax]
            ymax   = hi[yax]
            tex    = self.renderTexture

            self.drawCrossSection(zpos, axes, lo, hi, tex)

            tex.drawOnBounds(zpos, xmin, xmax, ymin, ymax, xax, yax, xform)


    def draw3D(self, xform=None, bbox=None):
        """Overrides :meth:`.GLObject.draw3D`. Draws a 3D rendering of the
        mesh.
        """
        opts      = self.opts
        verts     = self.vertices
        idxs      = self.indices
        normals   = self.normals
        blo, bhi  = self.getDisplayBounds()
        vdata     = self.getVertexData('vertex')
        mdata     = self.getVertexData('modulate')

        if mdata is None:
            mdata = vdata

        is2D = np.isclose(bhi[2], blo[2])

        if opts.wireframe:
            gl.glPolygonMode(gl.GL_FRONT_AND_BACK, gl.GL_LINE)
            gl.glLineWidth(2)
        else:
            gl.glPolygonMode(gl.GL_FRONT_AND_BACK, gl.GL_FILL)

        if xform is not None:
            gl.glMatrixMode(gl.GL_MODELVIEW)
            gl.glPushMatrix()
            gl.glMultMatrixf(np.array(xform, dtype=np.float32).ravel('F'))

        if is2D or opts.wireframe:
            enable = (gl.GL_DEPTH_TEST)
        else:
            enable = (gl.GL_DEPTH_TEST, gl.GL_CULL_FACE)

        gl.glDisable(gl.GL_CULL_FACE)
        with glroutines.enabled(enable):
            gl.glFrontFace(self.frontFace())
            if not is2D:
                gl.glCullFace(gl.GL_BACK)
            fslgl.glmesh_funcs.draw(
                self,
                gl.GL_TRIANGLES,
                verts,
                normals=normals,
                indices=idxs,
                vdata=vdata,
                mdata=mdata)

        if xform is not None:
            gl.glPopMatrix()


    def postDraw(self, xform=None, bbox=None):
        """Overrides :meth:`.GLObject.postDraw`. May call the
        :func:`.gl14.glmesh_funcs.postDraw` or
        :func:`.gl21.glmesh_funcs.postDraw` function.
        """

        if self.needShader():
            fslgl.glmesh_funcs.postDraw(self)

            if self.opts.vertexData is not None:
                if self.opts.useLut:
                    self.lutTexture.unbindTexture()
                else:
                    self.cmapTexture   .unbindTexture()
                    self.negCmapTexture.unbindTexture()


    def drawOutline(self, zpos, axes, xform=None, bbox=None):
        """Called by :meth:`draw2D` when ``MeshOpts.outline is True or
        MeshOpts.vertexData is not None``.  Calculates the intersection of the
        mesh with the viewing plane, and renders it as a set of
        ``GL_LINES``. If ``MeshOpts.vertexData is None``, the draw is
        performed using immediate mode OpenGL.

        Otherwise, the :func:`.gl14.glmesh_funcs.draw` or
        :func:`.gl21.glmesh_funcs.draw` function is used, which performs
        shader-based rendering.
        """

        opts = self.opts

        # Makes code below a bit nicer
        if xform is None:
            xform = np.eye(4)

        vertices, faces, dists, vertXform = self.calculateIntersection(
            zpos, axes, bbox)

        if vertXform is not None:
            xform = affine.concat(xform, vertXform)

        vdata     = self.getVertexData('vertex',   faces, dists)
        mdata     = self.getVertexData('modulate', faces, dists)
        useShader = vdata is not None
        vertices  = vertices.reshape(-1, 3)
        nvertices = vertices.shape[0]

        if mdata is None:
            mdata = vdata

        gl.glMatrixMode(gl.GL_MODELVIEW)
        gl.glPushMatrix()
        gl.glMultMatrixf(np.array(xform, dtype=np.float32).ravel('F'))

        gl.glLineWidth(opts.outlineWidth)

        # Constant colour
        if not useShader:
            vertices = vertices.ravel('C')
            gl.glColor(*opts.getConstantColour())
            gl.glEnableClientState(gl.GL_VERTEX_ARRAY)
            gl.glVertexPointer(3, gl.GL_FLOAT, 0, vertices)
            gl.glDrawArrays(gl.GL_LINES, 0, nvertices)
            gl.glDisableClientState(gl.GL_VERTEX_ARRAY)

        # Coloured from vertex data
        else:
            fslgl.glmesh_funcs.draw(
                self,
                gl.GL_LINES,
                vertices,
                vdata=vdata,
                mdata=mdata)

        gl.glPopMatrix()


    def draw2DMesh(self, xform=None, bbox=None):
        """Not to be confused with :meth:`draw2D`.  Called by :meth:`draw2D`
        for :class:`.Mesh` overlays which are actually 2D (with a flat
        third dimension).
        """

        opts      = self.opts
        vdata     = opts.getVertexData('vertex')
        mdata     = opts.getVertexData('modulate')
        useShader = self.needShader()
        vertices  = self.vertices
        faces     = self.indices

        if mdata is None:
            mdata = vdata

        if opts.outline:
            gl.glLineWidth(opts.outlineWidth)
            gl.glPolygonMode(gl.GL_FRONT_AND_BACK, gl.GL_LINE)
        else:
            gl.glPolygonMode(gl.GL_FRONT_AND_BACK, gl.GL_FILL)

        # Constant colour
        if not useShader:

            gl.glColor(*opts.getConstantColour())
            gl.glEnableClientState(gl.GL_VERTEX_ARRAY)

            gl.glVertexPointer(3, gl.GL_FLOAT, 0, vertices.ravel('C'))
            gl.glDrawElements(gl.GL_TRIANGLES,
                              faces.shape[0],
                              gl.GL_UNSIGNED_INT,
                              faces.ravel('C'))

            gl.glDisableClientState(gl.GL_VERTEX_ARRAY)

        # Coloured from vertex data
        else:
            # TODO separate modulateDataIndex?
            vdata = vdata[:, opts.vertexDataIndex]
            mdata = mdata[:, 0]
            fslgl.glmesh_funcs.draw(
                self,
                gl.GL_TRIANGLES,
                vertices,
                indices=faces,
                vdata=vdata,
                mdata=mdata)


    def drawCrossSection(self, zpos, axes, lo, hi, dest):
        """Renders a filled cross-section of the mesh to an off-screen
        :class:`.RenderTexture`. See:

        http://glbook.gamedev.net/GLBOOK/glbook.gamedev.net/moglgp/advclip.html

        :arg zpos: Position along the z axis
        :arg axes: Tuple containing ``(x, y, z)`` axis indices.
        :arg lo:   Tuple containing the low bounds on each axis.
        :arg hi:   Tuple containing the high bounds on each axis.
        :arg dest: The :class:`.RenderTexture` to render to.
        """

        opts     = self.opts
        xax      = axes[0]
        yax      = axes[1]
        zax      = axes[2]
        xmin     = lo[xax]
        ymin     = lo[yax]
        xmax     = hi[xax]
        ymax     = hi[yax]
        vertices = self.vertices.ravel('C')
        indices  = self.indices

        dest.bindAsRenderTarget()
        dest.setRenderViewport(xax, yax, lo, hi)

        # Figure out the equation of a plane
        # perpendicular to the Z axis, and
        # located at the z position. This is
        # used as a clipping plane to draw
        # the mesh intersection.
        clipPlaneVerts                = np.zeros((4, 3), dtype=np.float32)
        clipPlaneVerts[0, [xax, yax]] = [xmin, ymin]
        clipPlaneVerts[1, [xax, yax]] = [xmin, ymax]
        clipPlaneVerts[2, [xax, yax]] = [xmax, ymax]
        clipPlaneVerts[3, [xax, yax]] = [xmax, ymin]
        clipPlaneVerts[:,  zax]       =  zpos

        planeEq  = glroutines.planeEquation(clipPlaneVerts[0, :],
                                            clipPlaneVerts[1, :],
                                            clipPlaneVerts[2, :])

        gl.glClearColor(0, 0, 0, 0)

        gl.glClear(gl.GL_COLOR_BUFFER_BIT |
                   gl.GL_DEPTH_BUFFER_BIT |
                   gl.GL_STENCIL_BUFFER_BIT)

        gl.glEnableClientState(gl.GL_VERTEX_ARRAY)
        gl.glEnable(gl.GL_CLIP_PLANE0)
        gl.glEnable(gl.GL_CULL_FACE)
        gl.glEnable(gl.GL_STENCIL_TEST)
        gl.glFrontFace(gl.GL_CCW)
        gl.glPolygonMode(gl.GL_FRONT_AND_BACK, gl.GL_FILL)

        gl.glClipPlane(gl.GL_CLIP_PLANE0, planeEq)
        gl.glColorMask(gl.GL_FALSE, gl.GL_FALSE, gl.GL_FALSE, gl.GL_FALSE)

        # First and second passes - render front and
        # back faces separately. In the stencil buffer,
        # subtract the mask created by the second
        # render from the mask created by the first -
        # this gives us a mask which shows the
        # intersection of the mesh with the clipping
        # plane.
        gl.glStencilFunc(gl.GL_ALWAYS, 0, 0)
        direction = [gl.GL_INCR, gl.GL_DECR]

        # If the mesh coordinate transformation
        # has a negative determinant, it means
        # the back faces will be facing the camera,
        # so we need to render the back faces first.
        if npla.det(opts.getTransform('mesh', 'display')) > 0:
            faceOrder = [gl.GL_FRONT, gl.GL_BACK]
        else:
            faceOrder = [gl.GL_BACK,  gl.GL_FRONT]

        for face, direction in zip(faceOrder, direction):

            gl.glStencilOp(gl.GL_KEEP, gl.GL_KEEP, direction)
            gl.glCullFace(face)

            gl.glVertexPointer(3, gl.GL_FLOAT, 0, vertices)
            gl.glDrawElements(gl.GL_TRIANGLES,
                              len(indices),
                              gl.GL_UNSIGNED_INT,
                              indices)

        # Third pass - render the intersection
        # of the front and back faces from the
        # stencil buffer to the render texture.
        gl.glColorMask(gl.GL_TRUE, gl.GL_TRUE, gl.GL_TRUE, gl.GL_TRUE)

        gl.glDisable(gl.GL_CLIP_PLANE0)
        gl.glDisable(gl.GL_CULL_FACE)
        gl.glDisableClientState(gl.GL_VERTEX_ARRAY)

        gl.glStencilFunc(gl.GL_NOTEQUAL, 0, 255)

        gl.glColor(*opts.getConstantColour())

        # Disable alpha blending - we
        # just want the colour copied
        # into the texture as-is.
        with glroutines.disabled(gl.GL_BLEND):
            gl.glBegin(gl.GL_QUADS)
            gl.glVertex3f(*clipPlaneVerts[0, :])
            gl.glVertex3f(*clipPlaneVerts[1, :])
            gl.glVertex3f(*clipPlaneVerts[2, :])
            gl.glVertex3f(*clipPlaneVerts[3, :])
            gl.glEnd()

        gl.glDisable(gl.GL_STENCIL_TEST)

        dest.unbindAsRenderTarget()
        dest.restoreViewport()


    def calculateViewport(self, lo, hi, axes, bbox=None):
        """Called by :meth:`draw2D`. Calculates an appropriate viewport (the
        horizontal/vertical minimums/maximums in display coordinates) given
        the ``lo`` and ``hi`` ``GLMesh`` display bounds, and a display
        ``bbox``.
        """

        xax = axes[0]
        yax = axes[1]

        if bbox is not None and (lo[xax] < bbox[xax][0] or
                                 hi[xax] < bbox[xax][1] or
                                 lo[yax] > bbox[yax][0] or
                                 hi[yax] > bbox[yax][1]):

            xlen  = float(hi[xax] - lo[xax])
            ylen  = float(hi[yax] - lo[yax])
            ratio = xlen / ylen

            bblo = [ax[0] for ax in bbox]
            bbhi = [ax[1] for ax in bbox]

            lo   = [max(l, bbl) for l, bbl in zip(lo, bblo)]
            hi   = [min(h, bbh) for h, bbh in zip(hi, bbhi)]

            dxlen  = float(hi[xax] - lo[xax])
            dylen  = float(hi[yax] - lo[yax])

            dratio = dxlen / dylen

            if dratio > ratio:

                ndxlen   = xlen * (dylen / ylen)
                lo[xax] += 0.5 * (ndxlen - dxlen)
                hi[xax] -= 0.5 * (ndxlen - dxlen)

            elif dratio < ratio:

                ndylen   = ylen * (dxlen / xlen)
                lo[yax] += 0.5 * (ndylen - dylen)
                hi[yax] -= 0.5 * (ndylen - dylen)

        return lo, hi


    def calculateIntersection(self, zpos, axes, bbox=None):
        """Uses the :func:`.Mesh.planeIntersection` method to
        calculate the intersection of the mesh with the viewing plane at
        the given ``zpos``.

        :arg zpos:  Z axis coordinate at which the intersection is to be
                    calculated

        :arg axes:  Tuple containing the ``(x, y, z)`` axis indices.

        :arg bbox:  A tuple containing a ``([xlo, ylo, zlo], [xhi, yhi, zhi])``
                    bounding box to which the calculation can be restricted.

        :returns: A tuple containing:

                   - A ``(n, 2, 3)`` array which contains the two vertices of
                     a line for every intersected face (triangle) in the mesh.

                   - A ``(n, 3)`` array containing the intersected faces
                     (indices into the :attr:`.Mesh.vertices` array).

                   - A ``(n, 2, 3)`` array containing the barycentric
                     coordinates of the intersection line vertices.

                   - A ``(4, 4)`` array containing a transformation matrix
                     for transforming the line vertices into the display
                     coordinate system. May be ``None``, indicating that
                     no transformation is necessary.

        .. note:: The line vertices, and corresponding mesh triangle face
                  indices, which make up the cross section are saved
                  via the :meth:`.OverlayList.setData` method, with a key
                  ``'crosssection_[zax]'``, where ``[zax]`` is set to the
                  index of the display Z axis.
        """

        overlay     = self.overlay
        zax         = axes[2]
        opts        = self.opts
        origin      = [0] * 3
        normal      = [0] * 3
        origin[zax] = zpos
        normal[zax] = 1

        vertXform = opts.getTransform(           'mesh',    'display')
        origin    = opts.transformCoords(origin, 'display', 'mesh')
        normal    = opts.transformCoords(normal, 'display', 'mesh',
                                         vector=True)

        # TODO use bbox to constrain? This
        #      would be nice, but is not
        #      supported by trimesh.
        lines, faces, dists = overlay.planeIntersection(
            normal, origin, distances=True)

        lines = np.asarray(lines, dtype=np.float32)
        faces = np.asarray(faces, dtype=np.uint32)

        # cache the line vertices for other
        # things which might be interested.
        # See, for example, OrthoViewProfile
        # pick mode methods.
        self.overlayList.setData(overlay,
                                 'crosssection_{}'.format(zax),
                                 (lines, faces))

        faces = overlay.indices[faces]

        return lines, faces, dists, vertXform


    def getVertexData(self, vdtype, faces=None, dists=None):
        """If :attr:`.MeshOpts.vertexData` (or :attr:`.MeshOpts.modulateData`)
        is not ``None``, this method returns the vertex data to use for the
        current vertex data index.

        ``faces`` and ``dists`` are used by the :meth:`drawOutline` method,
        which draws a cross-section of the mesh. The ``dists`` array contains
        barycentric coordinates for each line vertex, and is used to linearly
        interpolate between the values of the vertices of the intersected
        triangles (defined in ``faces``).

        If ``MeshOpts.vertexData is None``, this method returns ``None``.

        :arg vdtype: Use ``'vertex'`` for :attr:`.MeshOpts.vertexData`, or
                     ``'modulate'`` for :attr:`.MeshOpts.modulateData`.
        """

        opts  = self.opts
        vdata = opts.getVertexData(vdtype)

        if vdata is None:
            return None

        # TODO separate modulateDataIndex for modulateData?
        if vdtype == 'vertex': vdata = vdata[:, opts.vertexDataIndex]
        else:                  vdata = vdata[:, 0]

        # when flat-shading, we colour each
        # triangle according to the data
        # value for its first vertex. This
        # should really be pre-generated
        # whenever the vertex data or flat-
        # shading option is changed, but it
        # is quick for typical surfaces, so
        # I'm not bothering for now.
        if self.threedee and (vdata is not None) and opts.flatShading:
            vdata = vdata[self.origIndices[:, 0]].repeat(3)

        if faces is not None and dists is not None:
            vdata = vdata[faces].repeat(2, axis=0).reshape(-1, 2, 3)
            vdata = (vdata * dists).reshape(-1, 3).sum(axis=1)

        return np.asarray(vdata, np.float32)


    def refreshCmapTextures(self, *a, **kwa):
        """Called when various :class:`.Display` or :class:`.MeshOpts``
        properties change. Refreshes the :class:`.ColourMapTexture` instances
        corresponding to the :attr:`.MeshOpts.cmap` and
        :attr:`.MeshOpts.negativeCmap` properties, and the
        :class:`.LookupTableTexture` corresponding to the :attr:`.MeshOpts.lut`
        property.

        :arg notify: Must be passed as a keyword argument. If ``True`` (the
                     default) :meth:`.GLObject.notify` is called after the
                     textures have been updated.
        """

        notify = kwa.pop('notify', True)

        display = self.display
        opts    = self.opts
        alpha   = display.alpha / 100.0
        cmap    = opts.cmap
        interp  = opts.interpolateCmaps
        res     = opts.cmapResolution
        negCmap = opts.negativeCmap
        gamma   = opts.realGamma(opts.gamma)
        invert  = opts.invert
        dmin    = opts.displayRange[0]
        dmax    = opts.displayRange[1]

        if interp: interp = gl.GL_LINEAR
        else:      interp = gl.GL_NEAREST

        self.cmapTexture.set(cmap=cmap,
                             invert=invert,
                             alpha=alpha,
                             resolution=res,
                             gamma=gamma,
                             interp=interp,
                             displayRange=(dmin, dmax))

        self.negCmapTexture.set(cmap=negCmap,
                                invert=invert,
                                alpha=alpha,
                                resolution=res,
                                gamma=gamma,
                                interp=interp,
                                displayRange=(dmin, dmax))

        self.lutTexture.set(alpha=display.alpha           / 100.0,
                            brightness=display.brightness / 100.0,
                            contrast=display.contrast     / 100.0,
                            lut=opts.lut)

        if notify:
            self.notify()


    def compileShaders(self):
        """(Re)Compiles the vertex/fragment shader program(s), via a call
        to the GL-version specific ``compileShaders`` function.
        """
        if self.flatShader is not None: self.flatShader.destroy()
        if self.dataShader is not None: self.dataShader.destroy()

        self.activeShader = None

        fslgl.glmesh_funcs.compileShaders(self)


    def updateShaderState(self):
        """Updates the vertex/fragment shader program(s) state, via a call to
        the GL-version specific ``updateShaderState`` function.
        """

        dopts      = self.opts
        canvas     = self.canvas
        flatColour = dopts.getConstantColour()
        useNegCmap = (not dopts.useLut) and dopts.useNegativeCmap

        if dopts.useLut:
            delta     = 1.0 / (dopts.lut.max() + 1)
            cmapXform = affine.scaleOffsetXform(delta, 0.5 * delta)
        else:
            cmapXform = self.cmapTexture.getCoordinateTransform()

        # calculate a scale+offset which transforms
        # modulate alpha value from the data range
        # into an alpha value, according to the
        # modulateRange
        modlo, modhi = dopts.modulateRange
        modRange     = modhi - modlo
        if modRange == 0:
            modScale  = 1
            modOffset = 0
        else:
            modScale  = 1 / modRange
            modOffset = -modlo / modRange

        fslgl.glmesh_funcs.updateShaderState(
            self,
            useNegCmap=useNegCmap,
            cmapXform=cmapXform,
            modScale=modScale,
            modOffset=modOffset,
            flatColour=flatColour)



#!/usr/bin/env python
#
# __init__.py - FSLeyes control panels.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""The ``controls`` package is the home of all *FSLeyes controls*, as
described in the :mod:`~fsleyes` package documentation.


Every control panel is a sub-class of either the :class:`.ControlPanel` or
:class:`.ControlToolBar`, and each panel allows the user to control some
aspect of either the specific :class:`.ViewPanel` that the control panel is
embedded within, or some general aspect of *FSLeyes*.
"""



#!/usr/bin/env python
#
# saveannotationsaction.py - Load/save annotations displayed on an OrthoPanel
#                            to file
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`LoadAnnotationsAction` and
:class:`SaveAnnotationsAction` classes, both FSLeyes actions which can be used
to save/load :mod:`.annotations` to/from a file.  This module is tightly
coupled to the implementations of the specific :class:`.AnnotationObject`
types that are supported:

 - :class:`.Point`
 - :class:`.Line`
 - :class:`.Arrow`
 - :class:`.Rect`
 - :class:`.Ellipse`
 - :class:`.TextAnnotation`


The ``SaveAnnotationsAction`` class is an action which is added to the
FSLeyes Tools menu, and which allows the user to save all annotations that
have been added to the canvases of an :class:`.OrthoPanel` to a file. This
file can then be loaded back in via the ``LoadAnnotationsAction``.


The logic for serialising and deserialising annotations to/from string
representations is also implemented in this module.


A FSLeyes annotations file is a plain text file where each line contains
a description of one :class:`.AnnotationObject`. An example of a FSLeyes
annotation file is::

    X Line lineWidth=4 colour=#0090a0 alpha=100 honourZLimits=False zmin=89.0  zmax=90.0  x1=48.7 y1=117.4 x2=39.1 y2=65.5
    Y Rect lineWidth=4 colour=#ff0800 alpha=100 honourZLimits=False zmin=107.0 zmax=108.0 filled=False border=True x=43.3 y=108.1 w=71.5 h=-62.7
    Y TextAnnotation lineWidth=5 colour=#e383ff alpha=100 honourZLimits=False zmin=107.0 zmax=108.0 text='' fontSize=20 x=10.9 y=154.2


Each line has the form::

    <canvas> <type> key=value [key=value ...]

where:

 - ``<canvas>`` is one of ``X``, ``Y`` or ``Z``, indicating the ortho canvas
   that the annotation is drawn on

 - ``<type>`` is the annotation type, one of ``Point``, ``Line``, ``Arrow``,
   ``Rect``, ``Ellipse`` or ``TextAnnotation``.

 - ``key=value`` contains the name and value of one property of the annotation.

The following key-value pairs are set for all annotation types:

 * ``colour`` - Annotation colour, as string of the form ``#RRGGBB``
 * ``lineWidth`` - Line width in pixels
 * ``alpha`` - Transparency, between 0 and 10
 * ``honourZLimits`` - ``True`` or ``False``, whether ``zmin`` and ``zmax``
   should be applied
 * ``zmin`` - Minimum depth value, as a floating point number
 * ``zmax`` - Maximum depth value, as a floating point number


The following additional key-value pairs are set for specific annotation
types.  All coordinates and lengths are relative to the display coordinate
system:

 * ``Point``
  * ``x`` X coordinate
  * ``y`` Y coordinate
 * ``Line`` and ``Arrow``
  * ``x1`` X coordinate of first point
  * ``y1`` Y coordinate of first point
  * ``x2`` X coordinate of second point
  * ``y2`` Y coordinate of second point
 * ``Rect`` and ``Ellipse``
  * ``filled`` ``True`` or ``False``, whether the rectangle/ellipse is
    filled
  * ``border`` ``True`` or ``False``, whether the rectangle/ellipse is
    drawn with a border
  * ``x`` X coordinate of one corner of the rectangle, or the ellipse centre
  * ``y`` Y coordinate of one corner of the rectangle, or the ellipse centre
  * ``w`` Rectangle width, relative to ``x``, or horizontal radius of elliipse
  * ``h`` Rectangle height, relative to ``y``, or vertical radius of elliipse
 * ``TextAnnotation``
  * ``text`` Displayed text, quoted with ``shlex.quote``
  * ``fontSize`` Font size in points (relative to the canvas scaling that
    was in place at the time that the text was created)
  * ``x`` Bottom left X coordinate of text
  * ``y`` Bottom left Y coordinate of text
"""


import os
import shlex
import logging

from typing import Dict, List

import wx

import fsleyes_widgets.utils.status as status
import fsl.utils.settings           as fslsettings
import fsleyes.views.orthopanel     as orthopanel
import fsleyes.actions              as actions
import fsleyes.strings              as strings
import fsleyes.gl.annotations       as annotations


log = logging.getLogger(__name__)


class SaveAnnotationsAction(actions.Action):
    """The ``SaveAnnotationsAction`` allos the user to save annotations
    that have been added to an :class:`.OrthoPanel` to a file.
    """


    @staticmethod
    def ignoreTool():
        """This action is not intended to be loaded as a FSLeyes plugin.
        Rather, it is used directly by the :class:`.AnnotationPanel` class.
        """
        return True


    @staticmethod
    def supportedViews():
        """This action is only intended to work with :class:`.OrthoPanel`
        views.
        """
        return [orthopanel.OrthoPanel]


    def __init__(self, overlayList, displayCtx, ortho):
        """Create a ``SaveAnnotationsAction``.

        :arg overlayList: The :class:`.OverlayList`
        :arg displayCtx:  The :class:`.DisplayContext`
        :arg ortho:       The :class:`.OrthoPanel`.
        """
        actions.Action.__init__(self, overlayList, displayCtx,
                                self.__saveAnnotations)
        self.__ortho = ortho


    def __saveAnnotations(self):
        """Show a dialog prompting the user for a file to save to,
        then serialises all annotations, and saves them to that file.
        """

        ortho   = self.__ortho
        msg     = strings.messages[self, 'saveFile']
        fromDir = fslsettings.read('loadSaveOverlayDir', os.getcwd())
        dlg     = wx.FileDialog(wx.GetApp().GetTopWindow(),
                                message=msg,
                                defaultDir=fromDir,
                                defaultFile='annotations.txt',
                                style=wx.FD_SAVE | wx.FD_OVERWRITE_PROMPT)

        if dlg.ShowModal() != wx.ID_OK:
            return

        annots  = {'X' : ortho.getXCanvas().getAnnotations().annotations,
                   'Y' : ortho.getYCanvas().getAnnotations().annotations,
                   'Z' : ortho.getZCanvas().getAnnotations().annotations}

        filePath = dlg.GetPath()
        errtitle = strings.titles[  self, 'saveFileError']
        errmsg   = strings.messages[self, 'saveFileError']

        with status.reportIfError(errtitle, errmsg, raiseError=False):
            with open(filePath, 'wt') as f:
                f.write(serialiseAnnotations(annots))


class LoadAnnotationsAction(actions.Action):
    """The ``LoadAnnotationsAction`` allos the user to load annotations
    from a file into an :class:`.OrthoPanel`.
    """


    @staticmethod
    def ignoreTool():
        """This action is not intended to be loaded as a FSLeyes plugin.
        Rather, it is used directly by the :class:`.AnnotationPanel` class.
        """
        return True


    @staticmethod
    def supportedViews():
        """This action is only intended to work with :class:`.OrthoPanel`
        views.
        """
        return [orthopanel.OrthoPanel]


    def __init__(self, overlayList, displayCtx, ortho):
        """Create a ``SaveAnnotationsAction``.

        :arg overlayList: The :class:`.OverlayList`
        :arg displayCtx:  The :class:`.DisplayContext`
        :arg ortho:       The :class:`.OrthoPanel`.
        """
        actions.Action.__init__(self, overlayList, displayCtx,
                                self.__loadAnnotations)
        self.__ortho = ortho


    def __loadAnnotations(self):
        """Show a dialog prompting the user for a file to load, then loads the
        annotations contained in the file and adds them to the
        :class:`OrthoPanel`.
        """

        ortho   = self.__ortho
        msg     = strings.messages[self, 'loadFile']
        fromDir = fslsettings.read('loadSaveOverlayDir', os.getcwd())
        dlg     = wx.FileDialog(wx.GetApp().GetTopWindow(),
                                message=msg,
                                defaultDir=fromDir,
                                style=wx.FD_OPEN)

        if dlg.ShowModal() != wx.ID_OK:
            return

        annots  = {'X' : ortho.getXCanvas().getAnnotations(),
                   'Y' : ortho.getYCanvas().getAnnotations(),
                   'Z' : ortho.getZCanvas().getAnnotations()}

        filePath = dlg.GetPath()
        errtitle = strings.titles[  self, 'loadFileError']
        errmsg   = strings.messages[self, 'loadFileError']
        with status.reportIfError(errtitle, errmsg, raiseError=False):
            with open(filePath, 'rt') as f:
                s = f.read().strip()
            allObjs = deserialiseAnnotations(s, annots)

            for canvas, objs in allObjs.items():
                annots[canvas].annotations.extend(objs)


def serialiseAnnotations(
        allAnnots : Dict[str, List[annotations.AnnotationObject]]) -> str:
    """Serialise all of the annotations for each canvas of an
    :class:`.OrthoPanel` to a string representation.

    :arg allAnnots: Dictionary where the keys are one of ``'X'``, ``'Y'`` or
                    ``'Z'``, and the values are lists of
                    :class:`.AnnotationObject` instances to be serialised.
    :returns:       String containing serialised annotations
    """

    serialised = []

    for canvas, annots in allAnnots.items():
        for obj in annots:
            serialised.append(serialiseAnnotation(obj, canvas))

    return '\n'.join(serialised)


def serialiseAnnotation(obj    : annotations.AnnotationObject,
                        canvas : str) -> str:
    """Convert the given :class:`.AnnotationObject` to a string representation.
    """

    # All properties on all annotation
    # types that are to be serialised
    keys = ['colour', 'lineWidth', 'alpha', 'honourZLimits', 'zmin', 'zmax',
            'filled', 'border', 'fontSize', 'coordinates', 'text', 'x', 'y',
            'w', 'h', 'x1', 'y1', 'x2', 'y2']

    # Custom formatters for some properties
    def formatColour(colour):
        colour = [int(round(c * 255)) for c in colour[:3]]
        return '#{:02x}{:02x}{:02x}'.format(*colour)

    formatters = {
        'colour' : formatColour,
        'text'   : shlex.quote
    }

    kvpairs = []
    for key in keys:
        val = getattr(obj, key, None)
        if val is not None:
            val = formatters.get(key, str)(val)
            kvpairs.append('{}={}'.format(key, val))

    otype = type(obj).__name__

    return '{} {} {}'.format(canvas, otype, ' '.join(kvpairs))


def deserialiseAnnotations(
        s      : str,
        annots : Dict[str, annotations.Annotations]
) -> Dict[str, List[annotations.AnnotationObject]]:
    """Deserialise all of the annotation specifications in the string ``s``,
    and create :class:`.AnnotationObject` instances from them. The
    ``AnnotationObject`` instances are created, but not added to the
    :class:`.Annotations`.

    :arg s:      String containing serialised annotations
    :arg annots: Dictionary where the keys are one of ``'X'``, ``'Y'`` or
                 ``'Z'``, and the values are the :class:`.Annotations`
                 instances for each :class:`.OrthoPanel` canvas.
    :returns:    Dictionary of ``{canvas : [AnnotationObject]}`` mappings.
    """

    objs = {'X' : [], 'Y' : [], 'Z' : []}

    for line in s.split('\n'):
        try:
            obj, canvas = deserialiseAnnotation(line, annots)
            objs[canvas].append(obj)
        except Exception as e:
            log.warning('Error parsing annotation (%s): %s ', e, line)

    return objs


def deserialiseAnnotation(
        s      : str,
        annots : Dict[str, annotations.Annotations]
) -> annotations.AnnotationObject:
    """Deserialises the annotation specification in the provided string,
    and creates an  :class:`.AnnotationObject` instance.

    :arg s:      String containing serialised annotation
    :arg annots: Dictionary where the keys are one of ``'X'``, ``'Y'`` or
                 ``'Z'``, and the values are the :class:`.Annotations`
                 instance for each :class:`.OrthoPanel` canvas.
    :returns:    Tuple containing:

                   - An :class:`.AnnotationObject` instance
                   - ``'X'``, ``'Y'`` or ``'Z'``, denoting the canvas that
                     the :class:`.AnnotationObject` is to be drawn on.
    """

    # Parser functions for some property types
    # (default for unlisted properties is float)
    def tobool(v):
        return v.lower() in 'true'

    parsers = {
        'colour'        : str,
        'honourZLimits' : tobool,
        'filled'        : tobool,
        'border'        : tobool,
        'text'          : str,
        'coordinates'   : str,
    }

    canvas, otype, kvpairs = s.strip().split(maxsplit=2)

    canvas  = canvas.upper()
    cls     = getattr(annotations, otype)
    kvpairs = dict([kv.split('=') for kv in shlex.split(kvpairs)])

    if canvas not in 'XYZ':
        raise ValueError('Canvas is not one of X, Y, '
                         'or Z ({})'.format(canvas))

    for k, v in kvpairs.items():
        kvpairs[k] = parsers.get(k, float)(v)

    return cls(annots[canvas], **kvpairs), canvas



print("import: 'fsleyes'")
import fsleyes

print("import: 'fsleyes.actions'")
import fsleyes.actions

print("import: 'fsleyes.controls'")
import fsleyes.controls

print("import: 'fsleyes.displaycontext'")
import fsleyes.displaycontext

print("import: 'fsleyes.editor'")
import fsleyes.editor

print("import: 'fsleyes.gl'")
import fsleyes.gl

print("import: 'fsleyes.plotting'")
import fsleyes.plotting

print("import: 'fsleyes.plugins'")
import fsleyes.plugins

print("import: 'fsleyes.profiles'")
import fsleyes.profiles

print("import: 'fsleyes.views'")
import fsleyes.views

